# -*- coding: utf-8 -*-

from kvirt import common
from kvirt.common import error, pprint, warning, get_ssh_pub_key
from kvirt.defaults import UBUNTUS, METADATA_FIELDS
import ovirtsdk4 as sdk
from ovirtsdk4 import Error as oerror
import ovirtsdk4.types as types
import os
import sys
from subprocess import call, check_output
import json
from time import sleep, time
import yaml
from http.client import HTTPSConnection
import ssl
from urllib.parse import urlparse
from random import choice
from string import ascii_lowercase
from shutil import which

oimages = {'CentOS-6-x86_64-GenericCloud.qcow2': 'CentOS 6 Generic Cloud Image',
           'CentOS-Atomic-Host-7-GenericCloud.qcow2': 'CentOS 7 Atomic Host Image',
           'CentOS-7-x86_64-GenericCloud.qcow2': 'CentOS 7 Generic Cloud Image',
           'cirros-0.4.0-x86_64-disk.img': 'CirrOS 0.4.0 for x86_64',
           'Fedora-Cloud-Base-24-1.2.x86_64.qcow2': 'Fedora 24 Cloud Base Image v20160921.0 for x86_64',
           'Fedora-Cloud-Base-25-1.3.x86_64.qcow2': 'Fedora 25 Cloud Base Image v20170106.0 for x86_64',
           'Fedora-Cloud-Base-26-1.5.x86_64.qcow2': 'Fedora 26 Cloud Base Image v1.5 for x86_64',
           'Fedora-Cloud-Base-27-1.6.x86_64.qcow2': 'Fedora 27 Cloud Base Image v1.6 for x86_64',
           'Fedora-Cloud-Base-28-1.1.x86_64.qcow2': 'Fedora 28 Cloud Base Image v1.1 for x86_64',
           'trusty-server-cloudimg-amd64-disk1.img':
           'Ubuntu Server 14.04 LTS (Trusty Tahr) Cloud Image v20170110 for x86_64',
           'xenial-server-cloudimg-amd64-disk1.img':
           'Ubuntu Server 16.04 LTS (Xenial Xerus) Cloud Image v20170111 for x86_64',
           'yakkety-server-cloudimg-amd64-disk1.img':
           'Ubuntu Server 16.10 (Yakkety Yak) Cloud Image v20170106 for x86_64'}


class KOvirt(object):
    def __init__(self, host='127.0.0.1', port=22, user='admin@internal',
                 password=None, insecure=True, ca_file=None, org=None, debug=False,
                 cluster='Default', datacenter='Default', ssh_user='root', filtervms=False, filteruser=False,
                 filtertag=None):
        try:
            url = f"https://{host}/ovirt-engine/api"
            self.conn = sdk.Connection(url=url, username=user,
                                       password=password, insecure=insecure,
                                       ca_file=ca_file)
        except oerror as e:
            error(f"Unexpected error: {e}")
            return None
        self.debug = debug
        self.vms_service = self.conn.system_service().vms_service()
        self.templates_service = self.conn.system_service().templates_service()
        self.sds_service = self.conn.system_service().storage_domains_service()
        self.datacenter = datacenter
        self.cluster = cluster
        self.host = host
        self.port = port
        self.user = user
        self.ca_file = ca_file
        self.org = org
        self.ssh_user = ssh_user
        self.filtervms = filtervms
        self.filteruser = filteruser
        self.filtertag = filtertag
        self.netprofiles = {}

    def close(self):
        self.api.disconnect()
        return

    def exists(self, name):
        vmsearch = self.vms_service.list(search=f'name={name}')
        if vmsearch:
            return True
        return False

    def net_exists(self, name):
        profiles_service = self.conn.system_service().vnic_profiles_service()
        if not self.netprofiles:
            for prof in profiles_service.list():
                networkinfo = self.conn.follow_link(prof.network)
                netdatacenter = self.conn.follow_link(networkinfo.data_center)
                if netdatacenter.name == self.datacenter:
                    self.netprofiles[prof.name] = prof.id
            if 'default' not in self.netprofiles:
                if 'ovirtmgmt' in self.netprofiles:
                    self.netprofiles['default'] = self.netprofiles['ovirtmgmt']
                elif 'rhevm' in self.netprofiles:
                    self.netprofiles['default'] = self.netprofiles['rhevm']
        if name in self.netprofiles:
            return True
        return False

    def disk_exists(self, pool, name):
        print("not implemented")

    def create(self, name, virttype=None, profile='', flavor=None, plan='kvirt',
               cpumodel='host-model', cpuflags=[], cpupinning=[], numcpus=2, memory=512,
               guestid='guestrhel764', pool='default', image=None,
               disks=[{'size': 10}], disksize=10, diskthin=True,
               diskinterface='virtio', nets=['default'], iso=None, vnc=True,
               cloudinit=True, reserveip=False, reservedns=False,
               reservehost=False, start=True, keys=[], cmds=[], ips=None,
               netmasks=None, gateway=None, nested=True, dns=None, domain=None,
               tunnel=False, files=[], enableroot=True, alias=[], overrides={},
               tags=[], storemetadata=False, sharedfolders=[], cmdline=None, placement=[], autostart=False,
               cpuhotplug=False, memoryhotplug=False, numamode=None, numa=[], pcidevices=[], tpm=False, rng=False,
               metadata={}, securitygroups=[], vmuser=None, guestagent=True):
        ip = None
        initialization = None
        memory = memory * 1024 * 1024
        clone = not diskthin
        custom_properties = []
        ignition = False
        if image is not None:
            templates_service = self.templates_service
            templateslist = templates_service.list()
            found = False
            for temp in templateslist:
                if temp.name == image:
                    if temp.memory > memory:
                        memory = temp.memory
                        pprint(f"Using {memory}Mb for memory as defined in template {image}")
                    _template = types.Template(name=image)
                    found = True
            if not found:
                return {'result': 'failure', 'reason': f"image {image} not found"}
            if image is not None and common.needs_ignition(image):
                ignition = True
                cloudinit = False
                ignitiondata = ''
                version = common.ignition_version(image)
                ignitiondata = common.ignition(name=name, keys=keys, cmds=cmds, nets=nets, gateway=gateway, dns=dns,
                                               domain=domain, files=files, enableroot=enableroot, overrides=overrides,
                                               version=version, plan=plan, compact=False, removetls=True, image=image,
                                               vmuser=vmuser)
                ignitiondata = ignitiondata.replace('\n', '')
                initialization = types.Initialization(custom_script=ignitiondata)
        else:
            _template = types.Template(name='Blank')
        ostype = 'rhcos_x64' if ignition else None
        _os = types.OperatingSystem(boot=types.Boot(devices=[types.BootDevice.HD, types.BootDevice.CDROM]), type=ostype)
        console = types.Console(enabled=True)
        description = [f"filter={self.filtertag}"] if self.filtertag is not None else []
        for entry in [field for field in metadata if field in METADATA_FIELDS]:
            description.append(f'{entry}={metadata[entry]}')
        description = ','.join(description)
        profiles_service = self.conn.system_service().vnic_profiles_service()
        if not self.netprofiles:
            for prof in profiles_service.list():
                networkinfo = self.conn.follow_link(prof.network)
                netdatacenter = self.conn.follow_link(networkinfo.data_center)
                if netdatacenter.name == self.datacenter:
                    self.netprofiles[prof.name] = prof.id
        cpu = types.Cpu(topology=types.CpuTopology(cores=numcpus, sockets=1))
        try:
            if placement:
                placement_policy = None
                vmhost = types.Host(name=choice(placement))
            else:
                placement_policy = None
                vmhost = None
            vm = self.vms_service.add(types.Vm(name=name, cluster=types.Cluster(name=self.cluster), memory=memory,
                                               cpu=cpu, description=description, template=_template, console=console,
                                               os=_os, placement_policy=placement_policy,
                                               custom_properties=custom_properties), clone=clone)
            vm_service = self.vms_service.vm_service(vm.id)
        except Exception as e:
            if self.debug:
                print(e)
            return {'result': 'failure', 'reason': e}
        if vnc:
            vnc_console = types.GraphicsConsole(protocol=types.GraphicsType.VNC)
            graphics_console_service = vm_service.graphics_consoles_service()
            graphical_consoles = graphics_console_service
            for gc in graphical_consoles.list():
                graphics_console_service.console_service(gc.id).remove()
            graphics_console_service.add(vnc_console)
        cdroms_service = vm_service.cdroms_service()
        cdrom = cdroms_service.list()[0]
        cdrom_service = cdroms_service.cdrom_service(cdrom.id)
        if iso is not None:
            iso_id = iso
            disks_service = self.conn.system_service().disks_service()
            disksearch = disks_service.list(search=f'name={iso}')
            if disksearch:
                iso_id = disksearch[0].id
            try:
                cdrom_service.update(cdrom=types.Cdrom(file=types.File(id=iso_id)))
            except:
                return {'result': 'failure', 'reason': f"Iso {iso} not found"}
        timeout = 0
        while True:
            vm = vm_service.get()
            if vm.status == types.VmStatus.DOWN:
                break
            else:
                timeout += 5
                sleep(5)
                pprint(f"Waiting for vm {name} to be ready")
            if timeout > 80:
                return {'result': 'failure', 'reason': 'timeout waiting for vm to be ready'}
        if 'default' not in self.netprofiles:
            if 'ovirtmgmt' in self.netprofiles:
                self.netprofiles['default'] = self.netprofiles['ovirtmgmt']
            elif 'rhevm' in self.netprofiles:
                self.netprofiles['default'] = self.netprofiles['rhevm']
        nics_service = self.vms_service.vm_service(vm.id).nics_service()
        currentnics = len(nics_service.list())
        nic_configurations = []
        for index, net in enumerate(nets):
            netname = None
            netmask = None
            mac = None
            if isinstance(net, str):
                netname = net
            elif isinstance(net, dict) and 'name' in net:
                netname = net['name']
                ip = None
                mac = net.get('mac')
                netmask = next((e for e in [net.get('mask'), net.get('netmask')] if e is not None), None)
                gateway = net.get('gateway')
                noconf = net.get('noconf', False)
                if not noconf and 'ip' in net:
                    ip = net['ip']
                # if 'alias' in net:
                #    alias = net['alias']
                if not noconf and ips and len(ips) > index and ips[index] is not None:
                    ip = ips[index]
                if not noconf and ip is not None and netmask is not None and gateway is not None:
                    nic_configuration = types.NicConfiguration(name=f'eth{index}', on_boot=True,
                                                               boot_protocol=types.BootProtocol.STATIC,
                                                               ip=types.Ip(version=types.IpVersion.V4, address=ip,
                                                                           netmask=netmask, gateway=gateway))
                    nic_configurations.append(nic_configuration)
            if netname is not None and netname in self.netprofiles:
                profile_id = self.netprofiles[netname]
                mac = types.Mac(address=mac)
                nic = types.Nic(name=f'eth{index}', mac=mac, vnic_profile=types.VnicProfile(id=profile_id))
                if index < currentnics:
                    currentnic = nics_service.list()[index]
                    currentnic_service = nics_service.nic_service(currentnic.id)
                    currentnic_service.update(nic=nic)
                else:
                    nics_service.add(nic)
        # disk_attachments_service = self.vms_service.vm_service(vm.id).disk_attachments_service()
        for index, disk in enumerate(disks):
            diskpool = pool
            diskthin = True
            disksize = 10
            if isinstance(disk, int):
                disksize = disk
            elif isinstance(disk, str) and disk.isdigit():
                disksize = int(disk)
            elif isinstance(disk, dict):
                disksize = disk.get('size', disksize)
                diskpool = disk.get('pool', pool)
                diskthin = disk.get('thin', diskthin)
            if index == 0 and image is not None:
                if disksize != 10:
                    self.update_image_size(vm.id, disksize)
                continue
            newdisk = self.add_disk(name, disksize, pool=diskpool, thin=diskthin)
            if newdisk['result'] == 'failure':
                return {'result': 'failure', 'reason': newdisk['reason']}
        if cloudinit and not custom_properties and initialization is None:
            custom_script = ''
            if storemetadata and overrides:
                storeoverrides = {k: overrides[k] for k in overrides if k not in ['password', 'rhnpassword',
                                                                                  'rhnactivationkey']}
                storedata = {'path': '/root/.metadata',
                             'content': yaml.dump(storeoverrides, default_flow_style=False, indent=2)}
                if files:
                    files.append(storedata)
                else:
                    files = [storedata]
            if files:
                data = common.process_files(files=files, overrides=overrides)
                if data != '':
                    custom_script += "write_files:\n"
                    custom_script += data
            cmds.insert(0, 'sed -i /192.168.122.1/d /etc/resolv.conf')
            cmds.insert(1, 'sleep 10')
            ignorednics = "docker0 tun0 tun1 cni0 flannel.1"
            gcmds = []
            if image is not None and image.lower().startswith('debian'):
                gcmds.append('echo "deb http://download.opensuse.org/repositories/home:/evilissimo:/deb/Debian_7.0/ ./"'
                             ' >> /etc/apt/sources.list')
                gcmds.append('gpg -v -a --keyserver http://download.opensuse.org/repositories/home:/evilissimo:/deb/'
                             'Debian_7.0/Release.key --recv-keys D5C7F7C373A1A299')
                gcmds.append('gpg --export --armor 73A1A299 | apt-key add -')
                gcmds.append('apt-get update')
                gcmds.append('apt-get -Y install ovirt-guest-agent')
                gcmds.append(f'sed -i "s/# ignored_nics.*/ignored_nics = {ignorednics}/" /etc/ovirt-guest-agent.conf')
                gcmds.append('service ovirt-guest-agent enable')
                gcmds.append('service ovirt-guest-agent restart')
            if image is not None and [x for x in UBUNTUS if x in image.lower()]:
                gcmds.append('echo deb http://download.opensuse.org/repositories/home:/evilissimo:/ubuntu:/16.04/'
                             'xUbuntu_16.04/ /')
                gcmds.append('wget http://download.opensuse.org/repositories/home:/evilissimo:/ubuntu:/16.04/'
                             'xUbuntu_16.04//Release.key')
                gcmds.append('apt-key add - < Release.key')
                gcmds.append('apt-get update')
                gcmds.append('apt-get -Y install ovirt-guest-agent')
                gcmds.append('sed -i "s/# ignored_nics.*/ignored_nics = docker0,tun0/" /etc/ovirt-guest-agent.conf')
            if gcmds:
                index = 1
                if image is not None and image.startswith('rhel'):
                    subindex = [i for i, value in enumerate(cmds) if value.startswith('subscription-manager')]
                    if subindex:
                        index = subindex.pop() + 1
                cmds = cmds[:index] + gcmds + cmds[index:]
            data = common.process_cmds(cmds=cmds, overrides=overrides)
            custom_script += "runcmd:\n"
            custom_script += data
            custom_script = None if custom_script == '' else custom_script
            vmuser = vmuser or 'root'
            user_name = common.get_user(image) if image is not None else vmuser
            root_password = None
            # dns_servers = '8.8.8.8 1.1.1.1'
            # dns_servers = dns if dns is not None else '8.8.8.8 1.1.1.1'
            publickeyfile = get_ssh_pub_key()
            if publickeyfile is not None:
                publickeyfile = open(publickeyfile).read().rstrip()
                if not keys:
                    keys = [publickeyfile]
                else:
                    keys.append(publickeyfile)
            keys = '\n'.join(keys)
            host_name = f"{name}.{domain}" if domain is not None else name
            initialization = types.Initialization(user_name=user_name, root_password=root_password,
                                                  authorized_ssh_keys=keys, host_name=host_name,
                                                  nic_configurations=nic_configurations, dns_servers=dns,
                                                  dns_search=domain, custom_script=custom_script)
        if start:
            vm_data = {'use_cloud_init': cloudinit, 'vm': types.Vm(initialization=initialization, host=vmhost)}
            if ignition:
                vm_data['use_ignition'] = True
            vm_service.start(**vm_data)
        if ip is not None:
            self.update_metadata(name, 'ip', ip)
        return {'result': 'success'}

    def start(self, name):
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vminfo = vmsearch[0]
        if str(vminfo.status) == 'down':
            vm = self.vms_service.vm_service(vmsearch[0].id)
            vm.start()
        return {'result': 'success'}

    def stop(self, name, soft=False):
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vminfo = vmsearch[0]
        if str(vminfo.status) != 'down':
            vm = self.vms_service.vm_service(vmsearch[0].id)
            vm.stop()
        return {'result': 'success'}

    def create_snapshot(self, name, base):
        vmsearch = self.vms_service.list(search=f'name={base}')
        if not vmsearch:
            error(f"VM {base} not found")
            return {'result': 'failure', 'reason': f"VM {base} not found"}
        vm = vmsearch[0]
        snapshots_service = self.vms_service.vm_service(vm.id).snapshots_service()
        snapshots_service.add(types.Snapshot(description=name))
        return

    def delete_snapshot(self, name, base):
        print("not implemented")
        return

    def list_snapshots(self, base):
        print("not implemented")
        return []

    def revert_snapshot(self, name, base):
        print("not implemented")
        return

    def restart(self, name):
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vm = vmsearch[0]
        status = str(vm.status)
        vm = self.vms_service.vm_service(vmsearch[0].id)
        if status == 'down':
            vm.start()
        else:
            vm.reboot()
        return {'result': 'success'}

    def info_host(self):
        data = {}
        api = self.conn.system_service().get()
        system_service = self.conn.system_service()
        data['version'] = api.product_info.version.full_version
        if api.summary.vms is not None:
            data['vms_running'] = api.summary.vms.total
        if api.summary.hosts is not None:
            data['hosts_summary'] = api.summary.hosts.total
        hosts = []
        hosts_service = self.conn.system_service().hosts_service()
        for host in hosts_service.list():
            hosts.append(host.name)
        data['hosts'] = hosts
        storage = []
        if api.summary.storage_domains is not None:
            data['storage_summary'] = api.summary.storage_domains.total
        sds_service = system_service.storage_domains_service()
        for sd in sds_service.list():
            storage.append(f"Storage Domain: {sd.name}")
        data['storage'] = storage
        return data

    def status(self, name):
        print("not implemented")
        return

    def list(self):
        vms = []
        system_service = self.conn.system_service()
        if self.filtertag is not None:
            vmslist = self.vms_service.list(search=f'description=plan*,filter={self.filtertag}*')
        elif self.filteruser:
            users_service = system_service.users_service()
            user_name = f'{self.user}-authz' if '@internal' in self.user else self.user
            userid = [u.id for u in users_service.list() if u.user_name == user_name][0]
            vmslist = self.vms_service.list(search=f'created_by_user_id={userid}')
        elif self.filtervms:
            vmslist = self.vms_service.list(search='description=plan=*,profile=*')
        else:
            vmslist = self.vms_service.list()
        for vm in vmslist:
            try:
                vms.append(self.info(vm.name, vm=vm))
            except:
                continue
        return sorted(vms, key=lambda x: x['name'])

    def console(self, name, tunnel=False, tunnelhost=None, tunnelport=22, tunneluser='root', web=False):
        connectiondetails = None
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vm = vmsearch[0]
        vm_service = self.vms_service.vm_service(vm.id)
        consoles_service = vm_service.graphics_consoles_service()
        consoles = consoles_service.list(current=True)
        for c in consoles:
            console_service = consoles_service.console_service(c.id)
            ticket = console_service.ticket()
            if str(c.protocol) == 'spice':
                ocacontent = open(self.ca_file).read().replace('\n', '\\n')
                try:
                    host = self.conn.follow_link(vm.host)
                    hostname = host.address
                except:
                    hostname = c.address
                subject = f'O={self.org},CN={hostname}'
                if tunnel:
                    localport1 = common.get_free_port()
                    localport2 = common.get_free_port()
                    command = "ssh -o LogLevel=QUIET -f -p %s -L %s:%s:%s -L %s:%s:%s %s@%s sleep 5"\
                        % (self.port, localport1, c.address, c.port, localport2, c.address, c.tls_port, self.ssh_user,
                           self.host)
                    os.system(command)
                address = '127.0.0.1' if tunnel else c.address
                port = localport1 if tunnel else c.port
                sport = localport2 if tunnel else c.tls_port
                connectiondetails = """[virt-viewer]
type=spice
host={address}
port={port}
password={ticket}
tls-port={sport}
fullscreen=0
title={name}:%d
enable-smartcard=0
enable-usb-autoshare=1
delete-this-file=1
usb-filter=-1,-1,-1,-1,0
tls-ciphers=DEFAULT
host-subject={subject}
ca={ocacontent}
toggle-fullscreen=shift+f11
release-cursor=shift+f12
secure-attention=ctrl+alt+end
secure-channels=main;inputs;cursor;playback;record;display;usbredir;smartcard""".format(subject=subject,
                                                                                        ocacontent=ocacontent,
                                                                                        address=address,
                                                                                        port=port,
                                                                                        sport=sport,
                                                                                        ticket=ticket.value,
                                                                                        name=name)
            elif str(c.protocol) == 'vnc':
                if tunnel:
                    localport1 = common.get_free_port()
                    command = "ssh -o LogLevel=QUIET -f -p %s -L %s:%s:%s %s@%s sleep 5"\
                        % (self.port, localport1, c.address, c.port, self.ssh_user, self.host)
                    os.system(command)
                address = '127.0.0.1' if tunnel else c.address
                port = localport1 if tunnel else c.port
                connectiondetails = """[virt-viewer]
type=vnc
host={address}
port={port}
password={ticket}
title={name}:%d
delete-this-file=1
toggle-fullscreen=shift+f11
release-cursor=shift+f12""".format(address=address, port=port, ticket=ticket.value, name=name)
        if connectiondetails is None:
            error(f"Couldn't retrieve connection details for {name}")
            sys.exit(1)
        if web:
            return "%s://%s:%s+%s" % (c.protocol, address, sport if str(c.protocol) == 'spice' else port, ticket.value)
        with open("/tmp/console.vv", "w") as f:
            f.write(connectiondetails)
        if self.debug or os.path.exists("/i_am_a_container"):
            msg = f"Use remote-viewer with this:\n{connectiondetails}" if not self.debug else connectiondetails
            pprint(msg)
        elif os.path.exists('/Users') and str(c.protocol) == 'vnc':
            pprint(f"Use password {ticket.value}")
            os.popen(f"open vnc://{address}:{port}")
        else:
            os.popen("remote-viewer /tmp/console.vv &")
        return

    def serialconsole(self, name, web=False):
        system_service = self.conn.system_service()
        users_service = system_service.users_service()
        user = users_service.list(search=f'usrname={self.user}-authz')[0]
        user_service = users_service.user_service(user.id)
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vm = vmsearch[0]
        permissions_service = self.vms_service.vm_service(vm.id).permissions_service()
        permissions_service.add(types.Permission(user=types.User(id=user.id), role=types.Role(name='UserVmManager')))
        keys_service = user_service.ssh_public_keys_service()
        publickeyfile = get_ssh_pub_key()
        if publickeyfile is None:
            error("neither id_rsa, id_dsa nor id_ed25519 public keys found in your .ssh directory. This is required")
            return
        publickeyfile = open(publickeyfile).read().rstrip()
        try:
            keys_service.add(key=types.SshPublicKey(content=publickeyfile))
        except:
            pass
        command = f"ssh -t -p 2222 ovirt-vmconsole@{self.host} connect --vm-name {name}"
        if web:
            return command
        call(command, shell=True)
        return

    def dnsinfo(self, name):
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            return None, None
        vm = vmsearch[0]
        dnsclient, domain = None, None
        description = vm.description.split(',')
        for description in vm.description.split(','):
            desc = description.split('=')
            if len(desc) == 2:
                if desc[0] == 'dnsclient':
                    dnsclient = desc[1]
                if desc[0] == 'domain':
                    domain = desc[1]
        return dnsclient, domain

    def info(self, name, vm=None, debug=False):
        conn = self.conn
        minimal = False
        if vm is None:
            vmsearch = self.vms_service.list(search=f'name={name}')
            if not vmsearch:
                error(f"VM {name} not found")
                return {}
            vm = vmsearch[0]
        else:
            minimal = True
        status = str(vm.status)
        yamlinfo = {'name': vm.name, 'disks': [], 'nets': [], 'status': status, 'instanceid': vm.id}
        template = conn.follow_link(vm.template)
        image = template.name
        yamlinfo['image'] = image
        yamlinfo['user'] = common.get_user(image)
        for description in vm.description.split(','):
            desc = description.split('=')
            if len(desc) == 2:
                if desc[0] == 'filter':
                    continue
                else:
                    yamlinfo[desc[0]] = desc[1]
        try:
            if status == 'up':
                host = conn.follow_link(vm.host)
                yamlinfo['host'] = host.name
        except:
            pass
        yamlinfo['memory'] = int(vm._memory / 1024 / 1024)
        cpus = vm.cpu.topology.cores * vm.cpu.topology.sockets
        yamlinfo['numcpus'] = cpus
        yamlinfo['creationdate'] = vm._creation_time.strftime("%d-%m-%Y %H:%M")
        devices = self.vms_service.vm_service(vm.id).reported_devices_service().list()
        ips = []
        for device in devices:
            if device.ips:
                for ip in device.ips:
                    if str(ip.version) == 'v4' and ip.address not in ['172.17.0.1', '127.0.0.1']:
                        ips.append(ip.address)
        nics = self.vms_service.vm_service(vm.id).nics_service().list()
        profiles_service = self.conn.system_service().vnic_profiles_service()
        if ips:
            yamlinfo['ip'] = ips[0]
            if len(ips) > 1:
                yamlinfo['ips'] = ips
        if minimal:
            return yamlinfo
        if not self.netprofiles:
            self.netprofiles = {}
            for profile in profiles_service.list():
                self.netprofiles[profile.id] = profile.name
        for nic in nics:
            device = nic.name
            mac = nic.mac.address
            network = 'N/A'
            if nic.vnic_profile is not None and nic.vnic_profile.id in self.netprofiles:
                network = self.netprofiles[nic.vnic_profile.id]
            network_type = str(nic.interface)
            yamlinfo['nets'].append({'device': device, 'mac': mac, 'net': network, 'type': network_type})
        attachments = self.vms_service.vm_service(vm.id).disk_attachments_service().list()
        for attachment in attachments:
            disk = conn.follow_link(attachment.disk)
            storagedomain = conn.follow_link(disk.storage_domains[0]).name if disk.storage_domains else ''
            device = disk.name
            disksize = int(disk.provisioned_size / 2**30)
            diskformat = str(disk.format)
            drivertype = str(disk.content_type)
            path = disk.id
            yamlinfo['disks'].append({'device': device, 'size': disksize, 'format': diskformat, 'type': drivertype,
                                      'path': f"{storagedomain}/{path}"})
        if image is None and 'kubetype' in yamlinfo and yamlinfo['kubetype'] == 'openshift':
            yamlinfo['user'] = 'core'
        cdroms_service = self.vms_service.vm_service(vm.id).cdroms_service()
        cdroms = cdroms_service.list()
        if cdroms:
            cdrom = cdroms[0]
            if cdrom.file is not None:
                iso = cdrom.file.id
                disks_service = self.conn.system_service().disks_service()
                disksearch = disks_service.list(search=f'id={iso}')
                if disksearch:
                    yamlinfo['iso'] = disksearch[0].name
        if debug:
            yamlinfo['debug'] = vars(vm)
        return yamlinfo

    def ip(self, name):
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return None
        vm = vmsearch[0]
        ips = []
        devices = self.vms_service.vm_service(vm.id).reported_devices_service().list()
        for device in devices:
            if device.ips:
                for i in device.ips:
                    if str(i.version) == 'v4' and i.address not in ['172.17.0.1', '127.0.0.1']:
                        ips.append(i.address)
        if not ips:
            return None
        else:
            return ips[-1]

    def volumes(self, iso=False):
        if iso:
            isos = []
            for pool in self.conn.system_service().storage_domains_service().list():
                sd_service = self.sds_service.storage_domain_service(pool.id)
                if str(pool.type) == 'iso':
                    file_service = sd_service.files_service()
                    for isofile in file_service.list():
                        isos.append(isofile._name)
                else:
                    disks_service = sd_service.disks_service()
                    for disk in disks_service.list():
                        if disk.name.endswith('.iso'):
                            isos.append(disk.name)
            return isos
        else:
            images = []
            templates_service = self.templates_service
            templateslist = templates_service.list()
            for template in templateslist:
                if template.name != 'Blank':
                    images.append(template.name)
            return images

    def delete(self, name, snapshots=False):
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vminfo = vmsearch[0]
        vm = self.vms_service.vm_service(vminfo.id)
        if str(vminfo.status) not in ['down', 'unknown']:
            vm.stop()
            while True:
                sleep(5)
                currentvm = vm.get()
                if currentvm.status == types.VmStatus.DOWN:
                    break
        vm.remove()
        return {'result': 'success'}

    def clone(self, old, new, full=False, start=False):
        print("not implemented")
        return

    def update_metadata(self, name, metatype, metavalue, append=False):
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return
        vminfo = vmsearch[0]
        found = False
        newdescription = []
        olddescription = vminfo.description.split(',')
        for index, description in enumerate(olddescription):
            desc = description.split('=')
            if len(desc) == 2 and desc[0] == metatype:
                found = True
                if append:
                    oldvalue = desc[1]
                    metavalue = f"{oldvalue}+{metavalue}"
                newdescription.append(f"{metatype}={metavalue}")
            else:
                newdescription.append(description)
        if not found:
            newdescription.append(f"{metatype}={metavalue}")
        description = ','.join(newdescription)
        vm = self.vms_service.vm_service(vmsearch[0].id)
        vm.update(vm=types.Vm(description=description))
        return

    def update_memory(self, name, memory):
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vminfo = vmsearch[0]
        vm = self.vms_service.vm_service(vminfo.id)
        if str(vminfo.status) == 'up':
            warning("Note it will only be effective upon next start")
        memory = int(memory) * 1024 * 1024
        vm.update(vm=types.Vm(memory=memory))
        return {'result': 'success'}

    def update_cpus(self, name, numcpus):
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vminfo = vmsearch[0]
        if str(vminfo.status) == 'up':
            warning("Note it will only be effective upon next start")
        vm = self.vms_service.vm_service(vminfo.id)
        cpu = types.Cpu(topology=types.CpuTopology(cores=numcpus, sockets=1))
        vm.update(vm=types.Vm(cpu=cpu))
        return {'result': 'success'}

    def update_start(self, name, start=True):
        print("not implemented")
        return

    def update_information(self, name, information):
        self.update_metadata(name, 'information', information)
        return

    def update_iso(self, name, iso):
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vminfo = vmsearch[0]
        vm = self.vms_service.vm_service(vminfo.id)
        cdroms_service = vm.cdroms_service()
        cdrom = cdroms_service.list()[0]
        cdrom_service = cdroms_service.cdrom_service(cdrom.id)
        current = False if str(vminfo.status) == 'down' else True
        if iso is not None and iso != '':
            disks_service = self.conn.system_service().disks_service()
            disksearch = disks_service.list(search=f'name={iso}')
            if disksearch:
                iso_id = disksearch[0].id
            else:
                error(f"Iso {iso} not found")
                return {'result': 'failure', 'reason': f"Iso {iso} not found"}
            try:
                cdrom_service.update(cdrom=types.Cdrom(file=types.File(id=iso_id)), current=current)
            except Exception as e:
                error(f"Hit issue {e.fault._detail}")
                return {'result': 'failure', 'reason': f"Hit issue {e.fault._detail}"}
        else:
            cdrom_service.update(cdrom=types.Cdrom(file=types.File(id='')), current=current)
        return {'result': 'success'}

    def update_flavor(self, name, flavor):
        print("Not implemented")
        return {'result': 'success'}

    def create_disk(self, name, size, pool=None, thin=True, image=None):
        print("not implemented")
        return

    def add_disk(self, name, size, pool=None, thin=True, image=None,
                 shareable=False, existing=None, interface='virtio', novm=False, overrides={}, diskname=None):
        size *= 2**30
        system_service = self.conn.system_service()
        sds_service = system_service.storage_domains_service()
        poolcheck = sds_service.list(search=f'name={pool}')
        if not poolcheck:
            error(f"Pool {pool} not found")
            return {'result': 'failure', 'reason': f"Pool {pool} not found"}
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vm = self.vms_service.vm_service(vmsearch[0].id)
        disk_attachments_service = vm.disk_attachments_service()
        currentdisk = len(disk_attachments_service.list())
        diskindex = currentdisk + 1 if currentdisk > 0 else 0
        diskname = f'{name}_Disk{diskindex}'
        _format = types.DiskFormat.COW if thin else types.DiskFormat.RAW
        storagedomain = types.StorageDomain(name=pool)
        bootable = True if diskindex == 0 else False
        disk_attachment = types.DiskAttachment(disk=types.Disk(name=diskname, format=_format, provisioned_size=size,
                                                               storage_domains=[storagedomain]),
                                               interface=types.DiskInterface.VIRTIO, bootable=bootable, active=True)
        disk_attachment = disk_attachments_service.add(disk_attachment)
        disks_service = self.conn.system_service().disks_service()
        disk_service = disks_service.disk_service(disk_attachment.disk.id)
        timeout = 0
        while True:
            disk = disk_service.get()
            if disk.status == types.DiskStatus.OK:
                break
            else:
                timeout += 5
                sleep(5)
                pprint(f"Waiting for disk {diskname} to be ready")
            if timeout > 40:
                return {'result': 'failure', 'reason': f'timeout waiting for disk {diskname} to be ready'}
        return {'result': 'success'}

    def update_image_size(self, vmid, size):
        size *= 2**30
        vm = self.vms_service.vm_service(vmid)
        disk_attachments_service = vm.disk_attachments_service()
        diskid = disk_attachments_service.list()[0].id
        disk_attachment_service = disk_attachments_service.attachment_service(diskid)
        disk_attachment = types.DiskAttachment(disk=types.Disk(provisioned_size=size))
        disk_attachment_service.update(disk_attachment)
        disks_service = self.conn.system_service().disks_service()
        disk_service = disks_service.disk_service(diskid)
        timeout = 0
        while True:
            disk = disk_service.get()
            diskname = disk.name
            if disk.status == types.DiskStatus.OK:
                break
            else:
                timeout += 5
                sleep(5)
                pprint(f"Waiting for image disk {diskname} to be resized")
            if timeout > 40:
                return {'result': 'failure', 'reason': f'timeout waiting for image disk {diskname} to be resized'}
        return {'result': 'success'}

    def delete_disk(self, name=None, diskname=None, pool=None, novm=False):
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vm = self.vms_service.vm_service(vmsearch[0].id)
        disk_attachments_service = vm.disk_attachments_service()
        for disk in disk_attachments_service.list():
            if disk.id == diskname:
                disk_attachment_service = disk_attachments_service.attachment_service(disk.id)
                disk_attachment_service.update(types.DiskAttachment(active=False))
                while True:
                    sleep(5)
                    disk_attachment = disk_attachment_service.get()
                    if not disk_attachment.active:
                        break
                disks_service = self.conn.system_service().disks_service()
                disk_service = disks_service.disk_service(disk.disk.id)
                disk_service.remove()
                return {'result': 'success'}
        error(f"Disk {diskname} not found")
        return {'result': 'failure', 'reason': f"Disk {diskname} not found"}

    def list_disks(self):
        volumes = {}
        disks_service = self.conn.system_service().disks_service()
        for disk in disks_service.list():
            diskname = disk._id
            if disk._storage_domains:
                pool = self.conn.follow_link(disk._storage_domains[0])._name
                path = disk._alias
                volumes[diskname] = {'pool': pool, 'path': path}
        return volumes

    def detach_disks(self, name):
        pprint(f"Detaching non primary disks from {name}")
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vm = self.vms_service.vm_service(vmsearch[0].id)
        disk_attachments_service = vm.disk_attachments_service()
        all_disks = disk_attachments_service.list()
        vm_disks = all_disks[1:] if len(all_disks) > 1 else []
        for disk in vm_disks:
            disk_attachment_service = disk_attachments_service.attachment_service(disk.id)
            disk_attachment_service.update(types.DiskAttachment(active=False))
            while True:
                sleep(5)
                disk_attachment = disk_attachment_service.get()
                if not disk_attachment.active:
                    break
        return {'result': 'success'}

    def add_nic(self, name, network, model='virtio'):
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vm = vmsearch[0]
        nics_service = self.vms_service.vm_service(vm.id).nics_service()
        index = len(nics_service.list())
        profiles_service = self.conn.system_service().vnic_profiles_service()
        netprofiles = {}
        for prof in profiles_service.list():
            networkinfo = self.conn.follow_link(prof.network)
            netdatacenter = self.conn.follow_link(networkinfo.data_center)
            if netdatacenter.name == self.datacenter:
                netprofiles[prof.name] = prof.id
        if 'default' not in netprofiles:
            if 'ovirtmgmt' in netprofiles:
                netprofiles['default'] = netprofiles['ovirtmgmt']
            elif 'rhevm' in netprofiles:
                netprofiles['default'] = netprofiles['rhevm']
        if network in netprofiles:
            profile_id = netprofiles[network]
            nics_service.add(types.Nic(name=f'eth{index}', vnic_profile=types.VnicProfile(id=profile_id)))
        else:
            return {'result': 'failure', 'reason': f"Network {network} not found"}
        return {'result': 'success'}

    def delete_nic(self, name, interface):
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vmid = vmsearch[0].id
        for nic in self.vms_service.vm_service(vmid).nics_service().list():
            if nic.name == interface:
                nics_service = self.vms_service.vm_service(vmid).nics_service()
                nic_service = nics_service.nic_service(nic.id)
                nic_service.remove()
                return {'result': 'success'}
        error(f"VM {name} not found")
        return {'result': 'failure', 'reason': f"VM {name} not found"}

    def create_pool(self, name, poolpath, pooltype='dir', user='qemu', thinpool=None):
        print("not implemented")
        return

    def delete_image(self, image, pool=None):
        if image.endswith('.iso'):
            pprint(f"Deleting Iso {image}")
            disks_service = self.conn.system_service().disks_service()
            disksearch = disks_service.list(search=f'name={image}')
            if not disksearch:
                return {'result': 'failure', 'reason': f"Image {image} not found"}
            else:
                disk_id = disksearch[0].id
                disk_service = disks_service.disk_service(disk_id)
                disk_service.remove()
                return {'result': 'success'}
        pprint("Deleting Template {image}")
        templates_service = self.templates_service
        templateslist = templates_service.list()
        for template in templateslist:
            if template.name == image:
                template_service = templates_service.template_service(template.id)
                template_service.remove()
                return {'result': 'success'}
        return {'result': 'failure', 'reason': f"Image {image} not found"}

    def add_image(self, url, pool, short=None, cmds=[], name=None, size=None, convert=False):
        downloaded = False
        shortimage = os.path.basename(url).split('?')[0]
        iso = True if shortimage.endswith('.iso') or name.endswith('.iso') else False
        if shortimage in self.volumes(iso=iso):
            pprint(f"Template {shortimage} already there")
            return {'result': 'success'}
        system_service = self.conn.system_service()
        profiles_service = self.conn.system_service().vnic_profiles_service()
        profile_id = None
        if not self.netprofiles:
            for prof in profiles_service.list():
                networkinfo = self.conn.follow_link(prof.network)
                netdatacenter = self.conn.follow_link(networkinfo.data_center)
                if netdatacenter.name == self.datacenter:
                    self.netprofiles[prof.name] = prof.id
        if 'default' not in self.netprofiles:
            if 'ovirtmgmt' in self.netprofiles:
                profile_id = self.netprofiles['ovirtmgmt']
            elif 'rhevm' in self.netprofiles:
                profile_id = self.netprofiles['rhevm']
        if profile_id is None:
            return {'result': 'failure', 'reason': "Couldn't find ovirtmgmt nor rhevm network!!!"}
        sds_service = system_service.storage_domains_service()
        poolcheck = sds_service.list(search=f'name={pool}')
        if not poolcheck:
            return {'result': 'failure', 'reason': f"Pool {pool} not found"}
        disks_service = self.conn.system_service().disks_service()
        disksearch = disks_service.list(search=f'alias={shortimage}')
        if not disksearch:
            if os.path.exists(url):
                pprint(f"Using {url} as path")
            elif not os.path.exists(f'/tmp/{shortimage}'):
                downloaded = True
                pprint(f"Downloading locally {shortimage}")
                downloadcmd = f"curl -Lo /tmp/{shortimage} -f '{url}'"
                code = os.system(downloadcmd)
                if code != 0:
                    return {'result': 'failure', 'reason': "Unable to download indicated image"}
            else:
                pprint(f"Using found /tmp/{shortimage}")
            BUF_SIZE = 128 * 1024
            image_path = os.path.abspath(url) if os.path.exists(url) else f'/tmp/{shortimage}'
            extensions = {'bz2': 'bunzip2 -f', 'gz': 'gunzip -f', 'xz': 'unxz -f', 'zst': 'zstd --decompress'}
            for extension in extensions:
                flag = '--decompress' if extension == 'zstd' else '-f'
                if shortimage.endswith(extension):
                    executable = extensions[extension]
                    if which(executable) is None:
                        pprint(f"{executable} not found. Can't uncompress image")
                        sys.exit(1)
                    else:
                        uncompresscmd = f"{executable} {flag} {image_path}"
                        os.system(uncompresscmd)
                        shortimage = shortimage.replace(f".{extension}", '')
                        image_path = f'/tmp/{shortimage}'
                        break
            image_size = os.path.getsize(image_path)
            if not iso:
                out = check_output(["qemu-img", "info", "--output", "json", image_path])
                image_info = json.loads(out)
                virtual_size = image_info["virtual-size"]
                content_type = types.DiskContentType.DATA
                disk_format = types.DiskFormat.COW
            else:
                virtual_size = image_size
                content_type = types.DiskContentType.ISO
                disk_format = types.DiskFormat.RAW
            sparse = True if disk_format == types.DiskFormat.COW else False
            disks_service = self.conn.system_service().disks_service()
            disk = disks_service.add(disk=types.Disk(name=os.path.basename(shortimage), content_type=content_type,
                                                     description='Kcli Uploaded disk', format=disk_format,
                                                     initial_size=image_size,
                                                     provisioned_size=virtual_size,
                                                     sparse=sparse,
                                                     storage_domains=[types.StorageDomain(name=pool)]))
            disk_service = disks_service.disk_service(disk.id)
            disk_id = disk.id
            while True:
                sleep(5)
                disk = disk_service.get()
                if disk.status == types.DiskStatus.OK:
                    break
            transfers_service = self.conn.system_service().image_transfers_service()
            transfer = transfers_service.add(types.ImageTransfer(image=types.Image(id=disk_id)))
            transfer_service = transfers_service.image_transfer_service(transfer.id)
            while transfer.phase == types.ImageTransferPhase.INITIALIZING:
                sleep(1)
                transfer = transfer_service.get()
            destination_url = urlparse(transfer.proxy_url)
            context = ssl._create_unverified_context()
            proxy_connection = HTTPSConnection(destination_url.hostname, destination_url.port, context=context)
            proxy_connection.putrequest("PUT", destination_url.path)
            proxy_connection.putheader('Content-Length', image_size)
            proxy_connection.endheaders()
            last_progress = time()
            pprint(f"Uploading image {shortimage}")
            with open(image_path, "rb") as disk:
                pos = 0
                while pos < image_size:
                    to_read = min(image_size - pos, BUF_SIZE)
                    chunk = disk.read(to_read)
                    if not chunk:
                        transfer_service.pause()
                        raise RuntimeError(f"Unexpected end of file at pos={pos}")
                    proxy_connection.send(chunk)
                    pos += len(chunk)
                    now = time()
                    if now - last_progress > 10:
                        pprint("Uploaded %.2f%%" % (float(pos) / image_size * 100))
                        last_progress = now
            response = proxy_connection.getresponse()
            if response.status != 200:
                transfer_service.pause()
                return {'result': 'failure', 'reason': f"Upload failed: {response.status} {response.reason}"}
            transfer_service.finalize()
            proxy_connection.close()
        else:
            disk_id = disksearch[0].id
            disk_service = disks_service.disk_service(disk_id)
        if iso:
            return {'result': 'success'}
        _template = types.Template(name='Blank')
        _os = types.OperatingSystem(boot=types.Boot(devices=[types.BootDevice.HD, types.BootDevice.CDROM]))
        console = types.Console(enabled=True)
        cpu = types.Cpu(topology=types.CpuTopology(cores=2, sockets=1))
        memory = 1024 * 1024 * 1024
        tempname = 'kcli_import' + ''.join(choice(ascii_lowercase) for _ in range(5))
        tempvm = self.vms_service.add(types.Vm(name=tempname, cluster=types.Cluster(name=self.cluster),
                                               memory=memory, cpu=cpu, template=_template, console=console, os=_os),
                                      clone=False)
        while True:
            pprint(f"Preparing temporary vm {tempname}")
            sleep(5)
            disk_service = disks_service.disk_service(disk_id)
            disk = disk_service.get()
            if disk.status == types.DiskStatus.OK:
                break
        tempvm_service = self.vms_service.vm_service(tempvm.id)
        _format = types.DiskFormat.COW
        storagedomain = types.StorageDomain(name=pool)
        disk_attachments = types.DiskAttachment(disk=types.Disk(id=disk_id, format=_format,
                                                                storage_domains=[storagedomain]),
                                                interface=types.DiskInterface.VIRTIO,
                                                bootable=True, active=True)
        disk_attachments_service = tempvm_service.disk_attachments_service()
        disk_attachments_service.add(disk_attachments)
        nics_service = self.vms_service.vm_service(tempvm.id).nics_service()
        nics_service.add(types.Nic(name='eth0', vnic_profile=types.VnicProfile(id=profile_id)))
        template = self.templates_service.add_from_vm(template=types.Template(name=shortimage, vm=tempvm))
        template_service = self.templates_service.template_service(template.id)
        while True:
            pprint("Converting temporary vm to template")
            sleep(5)
            template = template_service.get()
            if template.status == types.TemplateStatus.OK:
                break
        tempvmsearch = self.vms_service.list(search=f'name={tempname}')
        tempvminfo = tempvmsearch[0]
        tempvm = self.vms_service.vm_service(tempvminfo.id)
        tempvm.remove()
        if downloaded:
            os.remove(f'/tmp/{shortimage}')
        return {'result': 'success'}

    def create_network(self, name, cidr=None, dhcp=True, nat=True, domain=None, plan='kvirt', overrides={}):
        if 'vlan' not in overrides:
            return {'result': 'failure', 'reason': "Missing Vlan"}
        vlan = overrides['vlan']
        networks_service = self.conn.system_service().networks_service()
        networks_service.add(network=types.Network(name=name, data_center=types.DataCenter(name=self.datacenter),
                                                   vlan=types.Vlan(vlan), usages=[types.NetworkUsage.VM], mtu=1500))
        return

    def delete_network(self, name=None, cidr=None, force=False):
        print("not implemented")
        return

    def list_pools(self):
        return [pool.name for pool in self.conn.system_service().storage_domains_service().list()]

    def list_networks(self):
        networks = {}
        networks_service = self.conn.system_service().networks_service()
        for network in networks_service.list():
            networkname = network._name
            cidr = 'N/A'
            vlan = network._vlan
            if vlan is not None:
                cidr = f"vlan {vlan.id}"
            dhcp = network._id
            domainname = network._data_center
            domainname = self.conn.follow_link(network._data_center).name
            mode = network._description
            networks[networkname] = {'cidr': cidr, 'dhcp': dhcp, 'domain': domainname, 'type': 'routed', 'mode': mode}
        return networks

    def info_network(self, name):
        networkinfo = common.info_network(self, name)
        return networkinfo

    def list_subnets(self):
        print("not implemented")
        return {}

    def delete_pool(self, name, full=False):
        print("not implemented")
        return

    def network_ports(self, name):
        print("not implemented")
        return

    def vm_ports(self, name):
        return []

# returns the path of the pool, if it makes sense. used by kcli list --pools
    def get_pool_path(self, pool):
        poolsearch = self.conn.system_service().storage_domains_service().list(search=f'name={pool}')
        if not poolsearch:
            error(f"Pool {pool} not found")
            return {'result': 'failure', 'reason': f"Pool {pool} not found"}
        pool = poolsearch[0]
        return pool.storage.path

    def list_flavors(self):
        return []

    def export(self, name, image=None):
        vmsearch = self.vms_service.list(search=f'name={name}')
        if not vmsearch:
            error(f"VM {name} not found")
            return {'result': 'failure', 'reason': f"VM {name} not found"}
        vminfo = vmsearch[0]
        vm = self.vms_service.vm_service(vminfo.id)
        if str(vminfo.status) == 'up':
            vm.stop()
        attachments = self.conn.follow_link(vminfo.disk_attachments)
        disk_ids = [attachment.disk.id for attachment in attachments]
        _format = types.DiskFormat.COW
        attachments = [types.DiskAttachment(disk=types.Disk(id=disk_id, format=_format)) for disk_id in disk_ids]
        newvm = types.Vm(id=vminfo.id, disk_attachments=attachments)
        newname = image if image is not None else name
        template = types.Template(name=newname, vm=newvm)
        template = self.templates_service.add(template=template)
        template_service = self.templates_service.template_service(template.id)
        while True:
            sleep(5)
            template = template_service.get()
            if template.status == types.TemplateStatus.OK:
                break
        return {'result': 'success'}

    def get_hostname(self, address):
        conn = self.conn
        try:
            hosts_service = conn.system_service().hosts_service()
            for host in hosts_service.list():
                print(vars(host))
                break
            # host = hosts_service.list(search='name=myhost')[0]
            # host_service = hosts_service.host_service(host.id)
        except:
            return address

    def list_dns(self, domain):
        return []

    def create_bucket(self, bucket, public=False):
        print("not implemented")
        return

    def delete_bucket(self, bucket):
        print("not implemented")
        return

    def delete_from_bucket(self, bucket, path):
        print("not implemented")
        return

    def download_from_bucket(self, bucket, path):
        print("not implemented")
        return

    def upload_to_bucket(self, bucket, path, overrides={}, temp_url=False, public=False):
        print("not implemented")
        return

    def list_buckets(self):
        print("not implemented")
        return []

    def list_bucketfiles(self, bucket):
        print("not implemented")
        return []

    def openshift_installer_data(self, pool):
        clusters = self.conn.system_service().clusters_service().list(search=f'name={self.cluster}')
        if clusters:
            clusterid = clusters[0].id
        pools = self.conn.system_service().storage_domains_service().list(search=f'name={pool}')
        if pools:
            poolid = pools[0].id
        profiles_service = self.conn.system_service().vnic_profiles_service()
        for prof in profiles_service.list():
            networkinfo = self.conn.follow_link(prof.network)
            netdatacenter = self.conn.follow_link(networkinfo.data_center)
            if netdatacenter.name == self.datacenter:
                vnicid = prof.id
                break
        return clusterid, poolid, vnicid

    def reserve_dns(self, name, nets=[], domain=None, ip=None, alias=[], force=False, primary=False):
        print("not implemented")
        return

    def update_nic(self, name, index, network):
        print("not implemented")

    def update_network(self, name, dhcp=None, nat=None, domain=None, plan=None, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def list_security_groups(self, network=None):
        print("not implemented")
        return []

    def create_security_group(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def delete_security_group(self, name):
        print("not implemented")
        return {'result': 'success'}

    def update_security_group(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def info_subnet(self, name):
        print("not implemented")
        return {}

    def create_subnet(self, name, cidr, dhcp=True, nat=True, domain=None, plan='kvirt', overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def delete_subnet(self, name, force=False):
        print("not implemented")
        return {'result': 'success'}

    def update_subnet(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def list_dns_zones(self):
        print("not implemented")
        return []



from kvirt.config import Kconfig

network1, network2 = 'myvpc-eu', 'myvpc-us'
region1 = 'eu-west-3'
region2 = 'us-east-2'

config1 = Kconfig(region=region1)
config2 = Kconfig(region=region2)

vpc_network1 = config1.k.get_vpc_id(network1) if not network1.startswith('vpc-') else network1
vpc_network2 = config2.k.get_vpc_id(network2) if not network2.startswith('vpc-') else network2

Filters = [{'Name': "accepter-vpc-info.vpc-id", 'Values': [vpc_network2]},
           {'Name': "requester-vpc-info.vpc-id", 'Values': [vpc_network1]}]
for connection in config1.k.conn.describe_vpc_peering_connections(Filters=Filters)['VpcPeeringConnections']:
    config1.k.conn.delete_vpc_peering_connection(connection['VpcPeeringConnectionId'])



from kvirt.config import Kconfig
from time import sleep

network1, network2 = 'myvpc-eu', 'myvpc-us'
region1 = 'eu-west-3'
region2 = 'us-east-2'

config1 = Kconfig(region=region1)
config2 = Kconfig(region=region2)

vpc_network1 = config1.k.get_vpc_id(network1) if not network1.startswith('vpc-') else network1
vpc_network2 = config2.k.get_vpc_id(network2) if not network2.startswith('vpc-') else network2

# Filters = [{'Name': "accepter-vpc-info.vpc-id", 'Values': [vpc_network2]},
#           {'Name': "requester-vpc-info.vpc-id", 'Values': [vpc_network1]}]
# for connection in config1.k.conn.describe_vpc_peering_connections(Filters=Filters)['VpcPeeringConnections']:
#    print(connection)

response = config1.k.conn.create_vpc_peering_connection(VpcId=vpc_network1, PeerVpcId=vpc_network2, PeerRegion=region2)
peering_id = response['VpcPeeringConnection']['VpcPeeringConnectionId']
sleep(20)
config2.k.conn.accept_vpc_peering_connection(VpcPeeringConnectionId=peering_id)



from kvirt.config import Kconfig

cluster = 'testk'
network = "mynet"
api_ip = "12.0.0.251"
ingress_ip = "12.0.0.252"
cidr = "12.0.0.0/24"

config = Kconfig()
config.k.delete_network_port(f"{cluster}-api-vip")
config.k.delete_network_port(f"{cluster}-ingress-vip")
config.k.create_network(name=network, cidr=cidr, overrides={'port_security_enabled': True})
config.k.create_network_port(f"{cluster}-api-vip", network, ip=api_ip, floating=True)
config.k.create_network_port(f"{cluster}-ingress-vip", network, ip=ingress_ip, floating=True)



from kvirt.web import Kweb
import sys


def run():
    readonly = len(sys.argv) == 2 and sys.argv[1] == '--readonly'
    if readonly:
        print("Enabling readonly mode")
    web = Kweb(readonly=readonly)
    web.run()


if __name__ == '__main__':
    run()



import asyncio
from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client

from langchain_openai import ChatOpenAI
from langchain_mcp_adapters.tools import load_mcp_tools
from langgraph.prebuilt import create_react_agent

import sys

model = ChatOpenAI(model='granite3.2:latest', base_url='http://127.0.0.1:4000/v1', api_key='xxx')

server_params = StdioServerParameters(command="kmcp.py")


async def main_chat():
    async with stdio_client(server_params) as (read, write):
        async with ClientSession(read, write) as session:
            await session.initialize()
            tools = await load_mcp_tools(session)
            agent = create_react_agent(model, tools)

            print("Terminal chatbot is ready. Type 'exit' to quit.")
            chat_history = []

            while True:
                user_input = input("You: ").strip()
                if user_input.lower() in {"exit", "quit"}:
                    print("Exiting chatbot.")
                    break

                chat_history.append({"role": "user", "content": user_input})

                while True:
                    result = await agent.ainvoke({"messages": chat_history})
                    last_message = result["messages"][-1]

                    # Check if it's a tool_use message
                    if hasattr(last_message, 'type') and last_message.type == "tool_use":
                        # Agent wants to call a tool → MCP server handles it automatically
                        tool_name = last_message.tool_call[0]['name']
                        tool_args = last_message.tool_call[0]['args']
                        print(f"[Agent is calling tool '{tool_name}' with args {tool_args}]")

                        # Append tool_use message to chat history so agent remembers it
                        chat_history.append({"role": "user", "content": f"Calling tool: {tool_name}"})
                        continue  # Run agent again to get tool result processed

                    # Otherwise — final bot response
                    bot_reply = last_message.content
                    print(f"Bot: {bot_reply}")
                    chat_history.append({"role": "assistant", "content": bot_reply})
                    break


if __name__ == "__main__":
    try:
        asyncio.run(main_chat())
    except KeyboardInterrupt:
        print("\nChatbot interrupted. Exiting.")
        sys.exit(0)



from base64 import b64decode
import json
import os
from shutil import which
import ssl
from tempfile import NamedTemporaryFile
from urllib.request import urlopen, Request
import yaml


def _create_resource(kubectl, data, namespace=None, debug=False):
    if debug:
        print(f"CREATE with {data}")
    cmd = f"echo '{json.dumps(data)}' | {kubectl} create -o yaml -f -"
    if namespace is not None:
        cmd += f" -n {namespace}"
    return yaml.safe_load(os.popen(cmd).read())


def _delete_resource(kubectl, resource, name, namespace, debug=False):
    return os.popen(f'{kubectl} delete -n {namespace} {resource} {name} 2>/dev/null').read()


def _get_resource(kubectl, resource, name, namespace=None, debug=False):
    namespace = f" -n {namespace}" if namespace is not None else ''
    return yaml.safe_load(os.popen(f'{kubectl} get {resource} {namespace} -o yaml {name} 2>/dev/null').read())


def _get_all_resources(kubectl, resource, namespace=None, debug=False):
    namespace = f" -n {namespace}" if namespace is not None else ''
    cmd = f"{kubectl} get {resource} {namespace} -o yaml 2>/dev/null"
    return yaml.safe_load(os.popen(cmd).read())['items']


def _replace_resource(kubectl, data, namespace, debug=False):
    if 'status' in data:
        del data['status']
    if debug:
        print(f"REPLACE with {data}")
    return os.popen(f"echo '{json.dumps(data)}' | {kubectl} replace -n {namespace} -f -").read()


def _patch_resource(kubectl, resource, name, data, namespace, debug=False):
    if 'status' in data:
        del data['status']
    if debug:
        print(f"PATCH {resource} {name} with {data}")
    return os.popen(f"{kubectl} patch -n {namespace} {resource} {name} -p '{json.dumps(data)}'").read()


def _put(subresource, data, debug=False):
    headers = {'Content-Type': 'application/json'}
    kubectl = which('kubectl') or which('oc')
    cmd = '%s config view --minify --output jsonpath="{.clusters[*].cluster.server}"' % kubectl
    baseurl = os.popen(cmd).read()
    url = f'{baseurl}/{subresource}'
    cmd = "%s config view -o jsonpath='{.clusters[0]}' --raw" % kubectl
    ca_cert_data = json.loads(os.popen(cmd).read())['cluster']['certificate-authority-data']
    ca_cert_file = NamedTemporaryFile(delete=False)
    ca_cert_file.write(b64decode(ca_cert_data))
    ca_cert_file.close()
    context = ssl.create_default_context(cafile=ca_cert_file.name)
    cmd = "%s config view -o jsonpath='{.users[0]}' --raw" % kubectl
    kubeconfig_data = json.loads(os.popen(cmd).read())
    kubeconfig_user = kubeconfig_data['user']
    if 'client-certificate-data' in kubeconfig_user:
        client_cert_data = kubeconfig_user['client-certificate-data']
        client_cert_file = NamedTemporaryFile(delete=False)
        client_cert_file.write(b64decode(client_cert_data))
        client_cert_file.close()
        client_key_data = kubeconfig_user['client-key-data']
        client_key_file = NamedTemporaryFile(delete=False)
        client_key_file.write(b64decode(client_key_data))
        client_key_file.close()
        context.load_cert_chain(certfile=client_cert_file.name, keyfile=client_key_file.name)
    else:
        headers["Authorization"] = f"Bearer {kubeconfig_data['token']}"
    data = json.dumps(data).encode('utf-8')
    request = Request(url, headers=headers, method='PUT', data=data)
    urlopen(request, context=context)


def _create_secret(kubectl, name, namespace, data, field='userdata'):
    with NamedTemporaryFile(mode='w+') as tmp:
        tmp.write(data)
        tmp.flush()
        cmd = f"{kubectl} -n {namespace} create secret generic --from-file={field}={tmp.name} -o yaml {name}"
        return yaml.safe_load(os.popen(cmd).read())



#!/usr/bin/env python3
# coding=utf-8

from kvirt.config import Kconfig
from kvirt.common import get_user
import json
import sys
import argparse


def empty():
    return {'_meta': {'hostvars': {}}}


class KcliInventory(object):
    def __init__(self):
        self.inventory = {}
        self.read_cli_args()
        config = Kconfig(quiet=True)
        self.host = config.host
        self.port = config.port
        self.user = config.user
        self.tunnel = config.tunnel
        self.k = config.k
        self.type = config.type
        if self.k.conn is None:
            sys.exit(1)

        # Called with `--list`.
        if self.args.list:
            self.inventory = self._list()
        # Called with `--host [hostname]`.
        elif self.args.host:
            self.inventory = self.get(self.args.host)
        # If no groups or vars are present, return an empty inventory.
        else:
            self.inventory = empty()
        print(json.dumps(self.inventory))

    # Read the command line args passed to the script.
    def read_cli_args(self):
        parser = argparse.ArgumentParser()
        parser.add_argument('--list', action='store_true')
        parser.add_argument('--host', action='store')
        self.args = parser.parse_args()

    def _list(self):
        k = self.k
        tunnel = self.tunnel
        metadata = {'_meta': {'hostvars': {}}}
        hostvalues = metadata['_meta']['hostvars']
        for vm in k.list():
            name = vm.get('name')
            status = vm.get('status')
            ip = vm.get('ip', '')
            image = vm.get('image')
            plan = vm.get('plan', 'kvirt')
            if plan == '':
                plan = 'kvirt'
            profile = vm.get('profile', '')
            if plan not in metadata:
                metadata[plan] = {"hosts": [name], "vars": {"plan": plan, "profile": profile}}
            else:
                metadata[plan]["hosts"].append(name)
            hostvalues[name] = {'status': status}
            if tunnel and self.type in ['kvm', 'kubevirt']:
                hostvalues[name]['ansible_ssh_common_args'] = \
                    "-o ProxyCommand='ssh -p %s -W %%h:%%p %s@%s'" % (self.port, self.user, self.host)
            if ip != '':
                hostvalues[name]['ansible_host'] = ip
                if image is not None:
                    user = get_user(image)
                    hostvalues[name]['ansible_user'] = user
        return metadata

    def get(self, name):
        k = self.k
        tunnel = self.tunnel
        metadata = {}
        vm = k.info(name)
        if not vm:
            return metadata
        for entry in ['name', 'image', 'plan', 'profile', 'ip']:
            metadata[entry] = vm.get(entry)
        if metadata['plan'] == '':
            metadata['plan'] = 'kvirt'
        if tunnel and self.type in ['kvm', 'kubevirt']:
            metadata['ansible_ssh_common_args'] = \
                "-o ProxyCommand='ssh -p %s -W %%h:%%p %s@%s'" % (self.port, self.user, self.host)
        ip = metadata.get('ip')
        if ip is not None:
            metadata['ansible_host'] = ip
            image = metadata.get('image')
            if image is not None:
                user = get_user(image)
                metadata['ansible_user'] = user
        return metadata


def main():
    KcliInventory()


if __name__ == "__main__":
    main()



DHCPKEYWORDS = ['T1',
                'T2',
                'all-subnets-local',
                'arp-timeout',
                'boot-file-size',
                'bootfile-name',
                'bootfile-param',
                'bootfile-url',
                'classless-static-route',
                'client-arch',
                'client-interface-id',
                'client-machine-id',
                'default-ttl',
                'dns-server',
                'domain-name',
                'domain-search',
                'ethernet-encap',
                'extension-path',
                'information-refresh-time',
                'ip-forward-enable',
                'irc-server',
                'log-server',
                'lpr-server',
                'max-datagram-reassembly',
                'mobile-ip-home',
                'netbios-dd',
                'netbios-nodetype',
                'netbios-ns',
                'netbios-scope',
                'netmask',
                'nis+-domain',
                'nis+-server',
                'nis-domain',
                'nis-server',
                'nntp-server',
                'non-local-source-routing',
                'ntp-server',
                'policy-filter',
                'pop3-server',
                'root-path',
                'router',
                'router-discovery',
                'router-solicitation',
                'server-ip-address',
                'sip-server',
                'sip-server-domain',
                'smtp-server',
                'sntp-server',
                'static-route',
                'swap-server',
                'tcp-keepalive',
                'tcp-ttl',
                'tftp-server',
                'time-offset',
                'trailer-encapsulation',
                'user-class',
                'vendor-class',
                'vendor-id-encap',
                'x-windows-dm',
                'x-windows-fs']



#!/usr/bin/env python
# coding=utf-8

import argparse
from getpass import getuser
from jinja2 import Environment, FileSystemLoader
from jinja2.exceptions import TemplateSyntaxError, TemplateError
from kvirt.baseconfig import Kbaseconfig
from kvirt.cli import handle_parameters
from kvirt.common import pprint, error, container_mode
from kvirt.jinjafilters import jinjafilters
import os
import sys
from tempfile import TemporaryDirectory
import yaml

playbook_template = """---
- hosts: {{ hostname }}
  become: yes
  vars:
{% for override in overrides %}
    {{ override }}: {{ overrides[override] if overrides[override] != None else ''}}
{% endfor %}
  tasks:
{% for dir in dirs %}
  - name: Creates directory {{ dir }}
    file:
      path: {{ dir }}
      state: directory
{% endfor %}
{% for file in files %}
{% set mode = 700 if file['origin'].endswith('.py') or file['origin'].endswith('.sh') else 600 %}
  - name: Template file {{ file['origin'] }}
    {{ 'template' if file['render']|default(True) else 'copy' }}:
      src: {{ file['origin'] }}
      dest: {{ file['path'] }}
      owner: {{ file['owner']|default('root') }}
      mode: '{{ file['mode']|default(mode) }}'
{% endfor %}
{% if cmds != None %}
  - name: Copy cmds
    copy:
      content: {{ cmds }}
      dest: /root/.cmds.sh
      mode: '700'
  - name: Run cmds script
    shell: /root/.cmds.sh
    register: out
  - name: Result of cmds script
    debug: var=out.stdout_lines
{% endif %}
{% for script in scripts %}
  - name: Template script {{ script }}
    template:
      src: {{ script }}
      dest: /root/.{{ script }}
      owner: 'root'
      mode: '700'
  - name: Run script {{ script }} and log in /root/.{{ script }}.log
    shell: /root/.{{ script }} > /root/.{{ script }}.log 2>&1
#    register: out
#  - name: Result of script {{ script }}
#    debug: var=out.stdout_lines
{% endfor %}"""


def create_vm_playbook(env, name, profile, overrides={}):
    jinjadir = os.path.dirname(jinjafilters.__file__)
    if not os.path.exists('filter_plugins'):
        pprint("Creating symlink to kcli jinja filters")
        os.symlink(jinjadir, 'filter_plugins')
    dirs = []
    if 'scripts' not in profile:
        profile['scripts'] = []
    profile['cmds'] = '\n'.join(profile['cmds']) if 'cmds' in profile else None
    if 'files' in profile:
        files = []
        for _file in profile['files']:
            if isinstance(_file, str):
                entry = {'path': f'/root/{_file}', 'origin': _file, 'mode': 700}
            else:
                entry = _file
            if os.path.isdir(entry['origin']):
                dirs.append(entry['origin'])
                continue
            if entry['path'].count('/') > 2 and os.path.dirname(entry['path']) not in dirs:
                dirs.append(os.path.dirname(entry['path']))
            files.append(entry)
        profile['files'] = files
    else:
        profile['files'] = []
    try:
        templ = env.get_template(os.path.basename("playbook.j2"))
    except TemplateSyntaxError as e:
        error(f"Error rendering line {e.lineno} of file {e.filename}. Got: {e.message}")
        sys.exit(1)
    except TemplateError as e:
        error(f"Error rendering playbook. Got: {e.message}")
        sys.exit(1)
    hostname = overrides.get('hostname', name)
    profile['hostname'] = hostname
    if 'info' in overrides:
        del overrides['info']
    profile['overrides'] = overrides
    profile['dirs'] = dirs
    playbookfile = templ.render(**profile)
    playbookfile = '\n'.join([line for line in playbookfile.split('\n') if line.strip() != ""])
    print(playbookfile)


def create_playbook(baseconfig, plan, inputfile, overrides={}):
    playbooktmpdir = TemporaryDirectory()
    playbookdir = playbooktmpdir.name
    with open(f'{playbookdir}/playbook.j2', 'w') as f:
        f.write(playbook_template)
    env = Environment(loader=FileSystemLoader(playbookdir), extensions=['jinja2.ext.do'],
                      trim_blocks=True, lstrip_blocks=True)
    for jinjafilter in jinjafilters.jinjafilters:
        env.filters[jinjafilter] = jinjafilters.jinjafilters[jinjafilter]
    inputfile = os.path.expanduser(inputfile) if inputfile is not None else 'kcli_plan.yml'
    basedir = os.path.dirname(inputfile)
    if basedir == "":
        basedir = '.'
    pprint(f"Using plan {inputfile}...")
    pprint("Make sure to export ANSIBLE_JINJA2_EXTENSIONS=jinja2.ext.do")
    jinjadir = os.path.dirname(jinjafilters.__file__)
    if not os.path.exists('filter_plugins'):
        pprint("Creating symlink to kcli jinja filters")
        os.symlink(jinjadir, 'filter_plugins')
    if not os.path.exists(inputfile):
        error("No input file found nor default kcli_plan.yml. Leaving....")
        sys.exit(1)
    if 'baseplan' in overrides:
        del overrides['baseplan']
    entries, overrides, basefile, basedir = baseconfig.process_inputfile(plan, inputfile, overrides=overrides,
                                                                         full=True)
    config_data = {}
    config_data['config_host'] = baseconfig.ini[baseconfig.client].get('host', '127.0.0.1')
    config_data['config_type'] = config_data.get('config_type', 'kvm')
    default_user = getuser() if config_data['config_type'] == 'kvm'\
        and config_data['config_host'] in ['localhost', '127.0.0.1'] else 'root'
    config_data['config_user'] = config_data.get('config_user', default_user)
    overrides.update(config_data)
    renderfile = baseconfig.process_inputfile(plan, inputfile, overrides=overrides)
    try:
        data = yaml.safe_load(renderfile)
    except:
        error("Couldnt' parse plan. Leaving....")
        sys.exit(1)
    for key in data:
        if 'type' in data[key] and data[key]['type'] != 'kvm':
            continue
        elif 'scripts' not in data[key] and 'files' not in data[key] and 'cmds' not in data[key]:
            continue
        else:
            create_vm_playbook(env, key, data[key], overrides=overrides)
    playbooktmpdir.cleanup()


def run(args):
    plan = args.plan or 'playbook'
    overrides = handle_parameters(args.param, args.paramfile)
    inputfile = overrides.get('inputfile') or args.inputfile or 'kcli_plan.yml'
    if container_mode():
        inputfile = f"/workdir/{inputfile}"
    baseconfig = Kbaseconfig(client=args.client)
    _type = baseconfig.ini[baseconfig.client].get('type', 'kvm')
    overrides.update({'type': _type})
    create_playbook(baseconfig, plan, inputfile, overrides=overrides)


parser = argparse.ArgumentParser(description='Create playbook from plan')
parser.add_argument('-C', '-c', '--client')
parser.add_argument('-f', '--inputfile', help='Input Plan/File', default='kcli_plan.yml')
parser.add_argument('-P', '--param', action='append',
                    help='specify parameter or keyword for rendering (multiple can be specified)', metavar='PARAM')
parser.add_argument('--paramfile', '--pf', help='Parameters file', metavar='PARAMFILE', action='append')
parser.add_argument('plan', metavar='PLAN', nargs='?')

args = parser.parse_args()
run(args)



# -*- coding: utf-8 -*-

from ipaddress import ip_network
from kvirt import common
from kvirt.common import pprint, error, warning, get_ssh_pub_key
from kvirt.defaults import METADATA_FIELDS
from ibm_vpc import VpcV1, vpc_v1
import ibm_boto3
from ibm_botocore.client import Config
from ibm_cloud_sdk_core.authenticators import IAMAuthenticator
from ibm_cloud_sdk_core.api_exception import ApiException
from ibm_platform_services import GlobalTaggingV1, ResourceControllerV2, IamPolicyManagementV1, IamIdentityV1
from ibm_platform_services.iam_policy_management_v1 import PolicySubject, SubjectAttribute, PolicyResource, PolicyRole
from ibm_platform_services.iam_policy_management_v1 import ResourceAttribute
from ibm_cloud_networking_services import DnsRecordsV1, ZonesV1
import json
import os
from shutil import which
from time import sleep
from urllib.request import urlopen, Request
import webbrowser


def get_zone_href(region, zone):
    return f"https://{region}.iaas.cloud.ibm.com/v1/regions/{region}/zones/{zone}"


def get_s3_endpoint(region):
    return f'https://s3.{region}.cloud-object-storage.appdomain.cloud'


def get_service_instance_id(iam_api_key, name):
    if 'crn' in name:
        return name
    service_id = None
    headers = {'content-type': 'application/x-www-form-urlencoded', 'accept': 'application/json'}
    data = 'grant_type=urn%%3Aibm%%3Aparams%%3Aoauth%%3Agrant-type%%3Aapikey&apikey=%s' % iam_api_key
    data = data.encode()
    req = Request("https://iam.cloud.ibm.com/identity/token", headers=headers, method='POST', data=data)
    token = json.loads(urlopen(req).read().decode())['access_token']
    headers = {'Authorization': f'Bearer {token}'}
    req = Request("https://resource-controller.cloud.ibm.com/v2/resource_instances", headers=headers)
    for entry in json.loads(urlopen(req).read())['resources']:
        if entry['name'] == name:
            service_id = entry['id']
            break
    return service_id


class Kibm(object):
    """

    """
    def __init__(self, iam_api_key, region, zone, vpc, debug=False, cos_api_key=None, cos_resource_instance_id=None,
                 cis_resource_instance_id=None):
        self.debug = debug
        self.authenticator = IAMAuthenticator(iam_api_key)
        self.iam_api_key = iam_api_key
        self.conn = VpcV1(authenticator=self.authenticator)
        self.conn.set_service_url(f"https://{region}.iaas.cloud.ibm.com/v1")
        if cos_api_key is not None and cos_resource_instance_id is not None:
            cos_resource_instance_id = get_service_instance_id(iam_api_key, cos_resource_instance_id)
            self.s3 = ibm_boto3.client(
                's3',
                ibm_api_key_id=cos_api_key,
                ibm_service_instance_id=cos_resource_instance_id,
                ibm_auth_endpoint="https://iam.bluemix.net/oidc/token",
                config=Config(signature_version="oauth"),
                endpoint_url=get_s3_endpoint(region)
            )
            self.cos_resource_instance_id = cos_resource_instance_id
        self.global_tagging_service = GlobalTaggingV1(authenticator=self.authenticator)
        self.global_tagging_service.set_service_url('https://tags.global-search-tagging.cloud.ibm.com')
        if cis_resource_instance_id is not None:
            cis_resource_instance_id = get_service_instance_id(iam_api_key, cis_resource_instance_id)
            self.dns = ZonesV1(authenticator=self.authenticator, crn=cis_resource_instance_id)
            self.dns.set_service_url('https://api.cis.cloud.ibm.com')
            self.cis_resource_instance_id = cis_resource_instance_id
        self.resources = ResourceControllerV2(authenticator=self.authenticator)
        self.resources.set_service_url('https://resource-controller.cloud.ibm.com')
        self.iam_api_key = iam_api_key
        self.region = region
        self.zone = zone if region in zone else f"{region}-2"
        self.vpc = vpc

    def close(self):
        return

    def exists(self, name):
        try:
            return self._get_vm(name) is not None
        except ApiException as e:
            error(f"Unable to retrieve VM. Hit {e}")
            return False

    def net_exists(self, name):
        try:
            return self._get_subnet(name) is not None
        except ApiException as e:
            error(f"Unable to retrieve available subnets. Hit {e}")
            return False

    def disk_exists(self, pool, name):
        print("not implemented")

    def create(self, name, virttype=None, profile='', flavor=None, plan='kvirt',
               cpumodel='host-model', cpuflags=[], cpupinning=[], numcpus=2, memory=512,
               guestid='guestrhel764', pool='default', image=None,
               disks=[{'size': 10}], disksize=10, diskthin=True,
               diskinterface='virtio', nets=[], iso=None, vnc=True,
               cloudinit=True, reserveip=False, reservedns=False,
               reservehost=False, start=True, keys=[], cmds=[], ips=None,
               netmasks=None, gateway=None, nested=True, dns=None, domain=None,
               tunnel=False, files=[], enableroot=True, alias=[], overrides={},
               tags=[], storemetadata=False, sharedfolders=[], cmdline=None, cpuhotplug=False, memoryhotplug=False,
               numamode=None, numa=[], pcidevices=[], tpm=False, placement=[], autostart=False, rng=False, metadata={},
               securitygroups=[], vmuser=None, guestagent=True):
        try:
            vpcs = self.conn.list_vpcs().result['vpcs']
            for vpc in vpcs:
                if self.vpc == vpc['name']:
                    vpc_id = vpc['id']
                    resource_group_id = vpc['resource_group']['id']
                    break
            else:
                return {'result': 'failure', 'reason': f'VPC {self.vpc} does not exist'}
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to retrieve vpc information. Hit {e}'}
        if self.exists(name):
            return {'result': 'failure', 'reason': f"VM {name} already exists"}
        key_list = []
        try:
            pub_ssh_keys = [x['id'] for x in self.conn.list_keys().result['keys']]
            if not pub_ssh_keys:
                resource_group = {'id': resource_group_id}
                publickeyfile = get_ssh_pub_key()
                if publickeyfile is None:
                    return {'result': 'failure', 'reason': 'Unable to use a valid public ssh key'}
                pprint("Adding a default ssh public key named kvirt")
                identityfile = publickeyfile.replace('.pub', '')
                _type = identityfile.split('_')[-1]
                public_key = open(publickeyfile).read()
                new_key = self.conn.create_key(public_key, name='kvirt', resource_group=resource_group, type=_type)
                pub_ssh_keys.append(new_key.result['id'])
            for key in pub_ssh_keys:
                key_list.append(key)
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to check keys. Hit {e}'}
        if cloudinit:
            if image is not None and common.needs_ignition(image):
                version = common.ignition_version(image)
                userdata = common.ignition(name=name, keys=keys, cmds=cmds, nets=nets, gateway=gateway, dns=dns,
                                           domain=domain, files=files, enableroot=enableroot,
                                           overrides=overrides, version=version, plan=plan, image=image,
                                           vmuser=vmuser)
            else:
                userdata = common.cloudinit(name=name, keys=keys, cmds=cmds, nets=nets, gateway=gateway, dns=dns,
                                            domain=domain, files=files, enableroot=enableroot,
                                            overrides=overrides, fqdn=True, storemetadata=storemetadata,
                                            vmuser=vmuser)[0]
        else:
            userdata = ''
        if len(nets) == 0:
            return {'result': 'failure', 'reason': 'Network not found in configuration'}
        net_list = []
        subnets = {x['name']: x for x in self._get_subnets()}
        try:
            default_subnet = None
            subnets = {}
            for x in self._get_subnets():
                subnet_name = x['name']
                subnets[subnet_name] = x
                if x['vpc']['name'] == self.vpc and x['zone']['name'] == self.zone:
                    default_subnet = subnet_name
            for index, net in enumerate(nets):
                if isinstance(net, str):
                    netname = net
                elif isinstance(net, dict) and 'name' in net:
                    netname = net['name']
                if netname == 'default' or netname == self.vpc:
                    netname = default_subnet
                elif netname not in subnets:
                    return {'result': 'failure', 'reason': f'Network {netname} not found'}
                subnet = subnets[netname]
                if subnet['zone']['name'] != self.zone:
                    return {'result': 'failure', 'reason': f'Network {netname} is not in zone {self.zone}'}
                net_list.append(vpc_v1.NetworkInterfacePrototype(subnet=vpc_v1.SubnetIdentityById(id=subnet['id']),
                                                                 allow_ip_spoofing=False, name=f"eth{index}"))
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to check networks. Hit {e}'}
        if flavor is None:
            flavors = [f for f in self.list_flavors() if f[1] >= numcpus and f[2] * 1024 >= memory]
            if flavors:
                flavor = min(flavors, key=lambda f: f[2])[0]
                pprint(f"Using flavor {flavor}")
            else:
                return {'result': 'failure', 'reason': "Couldn't find a flavor matching cpu/memory requirements"}
        try:
            provisioned_profiles = self._get_profiles()
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to check flavors. Hit {e}'}
        if flavor not in provisioned_profiles:
            return {'result': 'failure', 'reason': f'Flavor {flavor} not found'}
        try:
            image = self._get_image(image)
            if image is None:
                return {'result': 'failure', 'reason': f'Image {image} not found'}
            image_id = image['id']
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to check provisioned images. Hit {e}'}
        volume_attachments = []
        for index, disk in enumerate(disks[1:]):
            disksize = int(disk.get('size')) if isinstance(disk, dict) and 'size' in disk else int(disk)
            diskname = f"{name}-disk{index + 1}"
            volume_by_capacity = {'capacity': disksize, 'name': diskname, 'profile': {'name': 'general-purpose'}}
            volume_attachment = {'delete_volume_on_instance_delete': True, 'volume': volume_by_capacity}
            volume_attachments.append(vpc_v1.VolumeAttachmentPrototypeInstanceContext.from_dict(volume_attachment))
        try:
            result_create = self.conn.create_instance(
                vpc_v1.InstancePrototypeInstanceByImage(
                    image=vpc_v1.ImageIdentityById(id=image_id),
                    primary_network_interface=net_list[0],
                    zone=vpc_v1.ZoneIdentityByHref(get_zone_href(self.region, self.zone)),
                    keys=[vpc_v1.KeyIdentityById(id=x) for x in key_list],
                    name=name,
                    network_interfaces=net_list[1:],
                    profile=vpc_v1.InstanceProfileIdentityByName(
                        name=flavor),
                    resource_group=vpc_v1.ResourceGroupIdentityById(id=resource_group_id),
                    volume_attachments=volume_attachments,
                    vpc=vpc_v1.VPCIdentityById(id=vpc_id),
                    user_data=userdata
                )
            ).result
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to create VM {name}. Hit {e}'}

        tag_names = []
        for entry in [field for field in metadata if field in METADATA_FIELDS]:
            tag_names.append(f'{entry}:{metadata[entry]}')
        resource_model = {'resource_id': result_create['crn']}
        try:
            self.global_tagging_service.attach_tag(resources=[resource_model],
                                                   tag_names=tag_names, tag_type='user').get_result()
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to attach tags. Hit {e}'}
        try:
            result_ip = self.conn.create_floating_ip(vpc_v1.FloatingIPPrototypeFloatingIPByTarget(
                target=vpc_v1.FloatingIPByTargetNetworkInterfaceIdentityNetworkInterfaceIdentityById(
                    id=result_create['network_interfaces'][0]['id']
                ),
                name=name,
                resource_group=vpc_v1.ResourceGroupIdentityById(
                    id=resource_group_id),
            )).result
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to create floating ip. Hit {e}'}
        try:
            self.conn.add_instance_network_interface_floating_ip(
                instance_id=result_create['id'],
                network_interface_id=result_create['network_interfaces'][0]['id'],
                id=result_ip['id']
            )
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to add floating ip. Hit {e}'}
        if reservedns and domain is not None:
            self.reserve_dns(name, nets=nets, domain=domain, alias=alias, instanceid=name)
        return {'result': 'success'}

    def start(self, name):
        try:
            vm = self._get_vm(name)
            if vm is None:
                return {'result': 'failure', 'reason': f'VM {name} not found'}
            vm_id = vm['id']
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to retrieve VM {name}. Hit {e}'}

        try:
            self.conn.create_instance_action(instance_id=vm_id, type='start')
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to start VM {name}. Hit {e}'}
        return {'result': 'success'}

    def stop(self, name, soft=False):
        try:
            vm = self._get_vm(name)
            if vm is None:
                return {'result': 'failure', 'reason': f'VM {name} not found'}
            vm_id = vm['id']
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to retrieve VM {name}. Hit {e}'}
        try:
            self.conn.create_instance_action(instance_id=vm_id, type='stop')
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to stop VM {name}. Hit {e}'}
        return {'result': 'success'}

    def create_snapshot(self, name, base):
        print("not implemented")
        return {'result': 'success'}

    def delete_snapshot(self, name, base):
        print("not implemented")
        return {'result': 'success'}

    def list_snapshots(self, base):
        print("not implemented")
        return []

    def revert_snapshot(self, name, base):
        print("not implemented")
        return {'result': 'success'}

    def restart(self, name):
        try:
            vm = self._get_vm(name)
            if vm is None:
                return {'result': 'failure', 'reason': f'VM {name} not found'}
            vm_id = vm['id']
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to retrieve VM {name}. Hit {e}'}
        try:
            self.conn.create_instance_action(instance_id=vm_id, type='reboot')
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to restart VM {name}. Hit {e}'}
        return {'result': 'success'}

    def info_host(self):
        data = {}
        data['region'] = self.region
        data['zone'] = self.zone
        data['vpc'] = self.vpc
        data['vms'] = len(self.list())
        return data

    def status(self, name):
        try:
            vm = self._get_vm(name)
            if vm is None:
                return {'result': 'failure', 'reason': f'VM {name} not found'}
            return vm['status']
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to retrieve VM {name}. Hit {e}'}

    def list(self):
        vms = []
        try:
            provisioned_vms = self._get_vms()
        except ApiException as e:
            error(f'Unable to retrieve VMs. Hit {e}')
            return vms
        try:
            floating_ips = {x['target']['id']: x for x in self.conn.list_floating_ips(
            ).result['floating_ips'] if x['status'] == 'available' and 'target' in x}
        except ApiException as e:
            error(f'Unable to retrieve floating ips. Hit {e}')
            return vms
        for vm in provisioned_vms:
            try:
                vms.append(self.info(vm['name'], vm=vm, ignore_volumes=True, floating_ips=floating_ips))
            except:
                continue
        return sorted(vms, key=lambda x: x['name'])

    def console(self, name, tunnel=False, tunnelhost=None, tunnelport=22, tunneluser='root', web=False):
        try:
            vm = self._get_vm(name)
            if vm is None:
                error(f"VM {name} not found")
                return None
        except ApiException as e:
            error(f"Unable to retrieve VM {name}. Hit {e}")
            return None
        try:
            url = f"https://cloud.ibm.com/vpc-ext/compute/vs/{self.region}~{vm['id']}/vnc"
        except ApiException as e:
            error(f"Unable to retrieve console access. Hit {e}")
            return None
        if web:
            return url
        if self.debug or os.path.exists("/i_am_a_container"):
            msg = f"Open the following url:\n{url}" if os.path.exists("/i_am_a_container") else url
            pprint(msg)
        else:
            pprint(f"Opening url: {url}")
            webbrowser.open(url, new=2, autoraise=True)
        return None

    def serialconsole(self, name, web=False):
        try:
            vm = self._get_vm(name)
            if vm is None:
                error(f"VM {name} not found")
                return None
        except ApiException as e:
            error(f"Unable to retrieve VM {name}. Hit {e}")
            return None
        try:
            url = f"https://cloud.ibm.com/vpc-ext/compute/vs/{self.region}~{vm['id']}/serial"
        except ApiException as e:
            error(f"Unable to retrieve console access. Hit {e}")
            return None
        if web:
            return url
        if self.debug or os.path.exists("/i_am_a_container"):
            msg = f"Open the following url:\n{url}" if os.path.exists("/i_am_a_container") else url
            pprint(msg)
        else:
            pprint(f"Opening url: {url}")
            webbrowser.open(url, new=2, autoraise=True)
        return None

    def info(self, name, output='plain', fields=[], values=False, vm=None, ignore_volumes=False, floating_ips=None,
             debug=False):
        yamlinfo = {}
        if vm is None:
            try:
                vm = self._get_vm(name)
                if vm is None:
                    error(f'VM {name} not found')
                    return yamlinfo
            except ApiException as e:
                error(f'Unable to retrieve VM {name}. Hit {e}')
                return yamlinfo
        state = vm['status']
        if floating_ips is None:
            try:
                floating_ips = {x['target']['id']: x for x in
                                self.conn.list_floating_ips().result['floating_ips'] if x['status'] == 'available'}
            except ApiException as e:
                error(f'Unable to retrieve floating ips. Hit {e}')
                return yamlinfo
        ips = []
        for network in vm['network_interfaces']:
            if network['id'] in floating_ips:
                ips.append(floating_ips[network['id']]['address'])
        for a in vm['network_attachments']:
            if 'virtual_network_interface' in a and a['virtual_network_interface']['id'] in floating_ips:
                ips.append(floating_ips[a['virtual_network_interface']['id']]['address'])
        ip = ','.join(ips)
        # zone = vm['zone']['name']
        image = vm['image']['name']
        yamlinfo['profile'] = vm['profile']['name']
        yamlinfo['name'] = name
        yamlinfo['status'] = state
        # yamlinfo['region'] = self.region
        # yamlinfo['zone'] = zone
        yamlinfo['ip'] = ip
        # yamlinfo['bandwidth'] = vm['bandwidth']
        yamlinfo['flavor'] = vm['profile']['name']
        yamlinfo['numcpus'] = vm['vcpu']['count']
        yamlinfo['memory'] = vm['memory']
        yamlinfo['image'] = image
        yamlinfo['user'] = common.get_user(image)
        yamlinfo['creationdate'] = vm['created_at']
        yamlinfo['id'] = vm['id']
        # yamlinfo['resource_group'] = vm['resource_group']['name']
        # yamlinfo['resource_type'] = vm['resource_type']
        # yamlinfo['startable'] = vm['startable']
        # yamlinfo['vpc'] = vm['vpc']['name']
        yamlinfo['profile'] = ''
        yamlinfo['plan'] = ''
        tag_list = self.global_tagging_service.list_tags(attached_to=vm['crn']).get_result().items()
        for entry in tag_list:
            if entry[0] != 'items':
                continue
            tags = entry[1]
            for tag in tags:
                tagname = tag['name']
                if tagname.count(':') == 1:
                    key, value = tagname.split(':')
                    if key in METADATA_FIELDS:
                        yamlinfo[key] = value
            break
        nets = []
        for interface in vm['network_interfaces']:
            network = interface['subnet']['name']
            device = interface['name']
            private_ip = interface['primary_ipv4_address']
            nets.append({'device': device, 'net': network, 'type': private_ip, 'mac': 'N/A'})
            yamlinfo['private_ip'] = private_ip
        if nets:
            yamlinfo['nets'] = nets
            # yamlinfo['primary_network_interface'] = vm['primary_network_interface']['name']
        disks = []
        if ignore_volumes is False:
            try:
                volumes = self._get_volumes()
            except ApiException as e:
                error(f"Unable to retrieve volume information. Hit {e}")
                return yamlinfo
            for attachment in vm['volume_attachments']:
                devname = attachment['volume']['name']
                if devname in volumes:
                    volume = volumes[devname]
                    disksize = volume['capacity']
                    drivertype = volume['profile']['name']
                    diskformat = 'N/A'
                    path = 'N/A'
                    disks.append({'device': devname, 'size': disksize, 'format': diskformat, 'type': drivertype,
                                  'path': path})
        if disks:
            yamlinfo['disks'] = disks
        if debug:
            yamlinfo['debug'] = vm
        return yamlinfo

    def ip(self, name):
        ips = []
        try:
            vm = self._get_vm(name)
            if vm is None:
                error(f'VM {name} not found')
                return ""
            for network in vm['network_interfaces']:
                response = self.conn.list_instance_network_interface_floating_ips(vm['id'], network['id'])
                ips.extend([x['address'] for x in response.result['floating_ips'] if x['status'] == 'available'])
        except ApiException as e:
            error(f"Unable to retrieve IP for {name}. Hit {e}")
        return ','.join(ips)

    def internalip(self, name):
        try:
            vm = self._get_vm(name)
        except ApiException:
            return None
        if 'primary_network_interface' not in vm:
            return None
        return vm['primary_network_interface']['primary_ipv4_address']

    def volumes(self, iso=False):
        image_list = []
        try:
            images = self.conn.list_images().result['images']
            for image in images:
                if image['status'] not in ['available', 'deprecated'] or \
                        image['operating_system']['name'].startswith('windows'):
                    continue
                image_list.append(image['name'])
        except ApiException as e:
            error(f"Unable to retrieve volume information. Hit {e}")
            return image_list
        return sorted(image_list, key=str.lower)

    def delete(self, name, snapshots=False):
        conn = self.conn
        try:
            vm = self._get_vm(name)
            if vm is None:
                return {'result': 'failure', 'reason': f'VM {name} not found'}
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to retrieve VM {name}. Hit {e}'}
        tags = []
        try:
            tags = self.global_tagging_service.list_tags(attached_to=vm['crn']).result['items']
        except Exception as e:
            error(f'Unable to retrieve tags. Hit {e}')
        dnsclient, domain = None, None
        for tag in tags:
            tagname = tag['name']
            if tagname.count(':') == 1:
                key, value = tagname.split(':')
                if key == 'domain':
                    domain = value
                if key == 'dnsclient':
                    dnsclient = value
        try:
            for network in vm['network_interfaces']:
                response = conn.list_instance_network_interface_floating_ips(instance_id=vm['id'],
                                                                             network_interface_id=network['id']).result
                if len(response['floating_ips']) == 0:
                    continue
                for floating_ip in response['floating_ips']:
                    conn.remove_instance_network_interface_floating_ip(id=floating_ip['id'],
                                                                       instance_id=vm['id'],
                                                                       network_interface_id=network['id'])
                    conn.delete_floating_ip(id=floating_ip['id'])
        except ApiException as e:
            return {'result': 'failure',
                    'reason': f'Unable to remove floating IPs for VM {name}. Hit {e}'}
        try:
            conn.delete_instance(id=vm['id'])
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to delete VM. Hit {e}'}
        if domain is not None and dnsclient is None:
            self.delete_dns(name, domain, name)
        return {'result': 'success'}

    def dnsinfo(self, name):
        dnsclient, domain = None, None
        try:
            vm = self._get_vm(name)
            if vm is None:
                error(f'VM {name} not found')
                return dnsclient, domain
        except ApiException as e:
            error(f'Unable to retrieve VM. Hit {e}')
            return dnsclient, domain
        try:
            tags = self.global_tagging_service.list_tags(attached_to=vm['crn']).result['items']
        except ApiException as e:
            error(f'Unable to retrieve tags. Hit {e}')
            return None, None
        for tag in tags:
            tagname = tag['name']
            if tagname.count(':') == 1:
                key, value = tagname.split(':')
                if key == 'dnsclient':
                    dnsclient = value
                if key == 'domain':
                    domain = value
        return dnsclient, domain

    def clone(self, old, new, full=False, start=False):
        print("not implemented")

    def update_metadata(self, name, metatype, metavalue, append=False):
        try:
            vm = self._get_vm(name)
            if vm is None:
                error(f'VM {name} not found')
                return
        except ApiException as e:
            error(f'Unable to retrieve VM {name}. Hit {e}')
            return
        resource_model = {'resource_id': vm['crn']}
        tag_names = [f"{metatype}:{metavalue}"]
        try:
            self.global_tagging_service.attach_tag(resources=[resource_model],
                                                   tag_names=tag_names, tag_type='user').get_result()
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to attach tags. Hit {e}'}

    def update_memory(self, name, memory):
        print("not implemented")

    def update_cpus(self, name, numcpus):
        print("not implemented")

    def update_start(self, name, start=True):
        print("not implemented")

    def update_information(self, name, information):
        self.update_metadata(name, 'information', information)

    def update_iso(self, name, iso):
        print("not implemented")

    def update_flavor(self, name, flavor):
        try:
            vm = self._get_vm(name)
            if vm is None:
                return {'result': 'failure', 'reason': f'VM {name} not found'}
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to retrieve VM {name}. Hit {e}'}
        if vm['status'] != 'stopped':
            return {'result': 'failure', 'reason': f'VM {name} must be stopped'}
        try:
            provisioned_profiles = self._get_profiles()
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to retrieve flavors. Hit {e}'}
        if flavor not in provisioned_profiles:
            return {'result': 'failure', 'reason': f'Flavor {flavor} not found'}
        try:
            self.conn.update_instance(id=vm['id'], instance_patch=vpc_v1.InstancePatch(
                profile=vpc_v1.InstancePatchProfileInstanceProfileIdentityByName(name=flavor)))
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to update instance. Hit {e}'}
        return {'result': 'success'}

    def create_disk(self, name, size, pool=None, thin=True, image=None):
        print("not implemented")

    def add_disk(self, name, size, pool=None, thin=True, image=None,
                 shareable=False, existing=None, interface='virtio', diskname=None):
        print("not implemented")

    def delete_disk(self, name, diskname, pool=None, novm=False):
        print("not implemented")

    def list_disks(self):
        print("not implemented")
        return {}

    def add_nic(self, name, network, model='virtio'):
        try:
            vm = self._get_vm(name)
            if vm is None:
                error(f'VM {name} not found')
                return
        except ApiException as e:
            error(f'Unable to retrieve VM {name}. Hit {e}')
            return
        try:
            subnet = self._get_subnet(network)
            if subnet is None:
                error(f'Network {network} not found')
                return
        except ApiException as e:
            error(f'Unable to retrieve network information. Hit {e}')
            return
        try:
            self.conn.create_instance_network_interface(
                instance_id=vm['id'],
                subnet=vpc_v1.SubnetIdentityById(id=subnet['id']),
                allow_ip_spoofing=False
            )
        except ApiException as e:
            error(f'Unable to create NIC. Hit {e}')

    def delete_nic(self, name, interface):
        try:
            vm = self._get_vm(name)
            if vm is None:
                error(f'VM {name} not found')
                return
        except ApiException as e:
            error(f'Unable to retrieve VM {name}. Hit {e}')
        try:
            for network in vm['network_interfaces']:
                if network['name'] == interface:
                    response = self.conn.delete_instance_network_interface(instance_id=vm['id'],
                                                                           id=network['id'])
                    if response.status_code != 204:
                        error(f'Unexpected status code received: {response.status_code}')
        except ApiException as e:
            error(f'Unable to delete NIC. Hit {e}')

    def create_pool(self, name, poolpath, pooltype='dir', user='qemu', thinpool=None):
        print("not implemented")

    def delete_image(self, image, pool=None):
        try:
            image = self._get_image(image)
            if image is None:
                return {'result': 'failure', 'reason': f'Image {image} not found'}
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to retrieve images. Hit {e}'}
        try:
            result = self.conn.delete_image(id=image['id'])
            if result.status_code != 202:
                return {'result': 'failure', 'reason': f'Unexpected status code received: {result.status_code}'}
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to delete image. Hit {e}'}
        return {'result': 'success'}

    def add_image(self, url, pool, short=None, cmds=[], name=None, size=None, convert=False):
        cos_id = self.cos_resource_instance_id.split(':')[7]
        identity_client = IamIdentityV1(authenticator=self.authenticator)
        api_key_detail = identity_client.get_api_keys_details(iam_api_key=self.iam_api_key).get_result()
        account_id = api_key_detail['account_id']
        iam_policy_management_service = IamPolicyManagementV1(authenticator=self.authenticator)
        image_policy_found = False
        for policy in iam_policy_management_service.list_policies(account_id).result['policies']:
            policy_type = policy['type']
            if policy_type != 'authorization':
                continue
            cos_found = [x for x in policy['resources'][0]['attributes'] if x['name'] == 'serviceInstance' and
                         x['value'] == cos_id]
            if cos_found:
                image_policy_found = True
                break
        if not image_policy_found:
            pprint("Adding authorization between image service and cloud storage instance")
            subject = PolicySubject(attributes=[SubjectAttribute(name='serviceName', value='is'),
                                                SubjectAttribute(name='accountId', value=account_id),
                                                SubjectAttribute(name='resourceType', value='image')])
            resources = PolicyResource(attributes=[ResourceAttribute(name='accountId', value=account_id),
                                                   ResourceAttribute(name='serviceName', value='cloud-object-storage'),
                                                   ResourceAttribute(name='serviceInstance', value=cos_id)])
            roles = [PolicyRole(role_id='crn:v1:bluemix:public:iam::::serviceRole:Writer')]
            iam_policy_management_service.create_policy(type='authorization', subjects=[subject],
                                                        roles=roles, resources=[resources]).get_result()
        if pool not in self.list_buckets():
            return {'result': 'failure', 'reason': f"Bucket {pool} doesn't exist"}
        shortimage = os.path.basename(url).split('?')[0]
        shortimage_unzipped = shortimage.replace('.gz', '')
        if shortimage_unzipped in self.volumes():
            return {'result': 'success'}
        delete_cos_image = False
        if shortimage_unzipped not in self.list_bucketfiles(pool):
            if not os.path.exists(f'/tmp/{shortimage}'):
                downloadcmd = f"curl -Lko /tmp/{shortimage} -f '{url}'"
                code = os.system(downloadcmd)
                if code != 0:
                    return {'result': 'failure', 'reason': "Unable to download indicated image"}
            if shortimage.endswith('gz'):
                if which('gunzip') is not None:
                    uncompresscmd = f"gunzip /tmp/{shortimage}"
                    os.system(uncompresscmd)
                else:
                    error("gunzip not found. Can't uncompress image")
                    return {'result': 'failure', 'reason': "gunzip not found. Can't uncompress image"}
                shortimage = shortimage_unzipped
            pprint("Uploading image to bucket")
            self.upload_to_bucket(pool, f'/tmp/{shortimage}')
            os.remove(f'/tmp/{shortimage}')
            delete_cos_image = True
        pprint("Importing image as template")
        image_file_prototype_model = {}
        image_file_prototype_model['href'] = f"cos://{self.region}/{pool}/{shortimage_unzipped}"
        operating_system_identity_model = {}
        operating_system_identity_model['name'] = 'centos-8-amd64'
        image_prototype_model = {}
        clean_image_name = shortimage_unzipped.replace('.', '-').replace('_', '-').lower()
        image_prototype_model['name'] = clean_image_name
        image_prototype_model['file'] = image_file_prototype_model
        image_prototype_model['operating_system'] = operating_system_identity_model
        image_prototype = image_prototype_model
        result_create = self.conn.create_image(image_prototype).get_result()
        while True:
            image = self._get_image(clean_image_name)
            if image['status'] == 'available':
                break
            else:
                pprint(f"Waiting for image {clean_image_name} to be available")
                sleep(10)
        tag_names = [f"image:{shortimage_unzipped}"]
        resource_model = {'resource_id': result_create['crn']}
        self.global_tagging_service.attach_tag(resources=[resource_model],
                                               tag_names=tag_names, tag_type='user').get_result()
        if delete_cos_image:
            self.delete_from_bucket(pool, shortimage_unzipped)
        return {'result': 'success'}

    def create_network(self, name, cidr=None, dhcp=True, nat=True, domain=None, plan='kvirt', overrides={}):
        if cidr is not None:
            try:
                network = ip_network(cidr)
            except:
                return {'result': 'failure', 'reason': f"Invalid Cidr {cidr}"}
            if str(network.version) == "6":
                return {'result': 'failure', 'reason': 'IPv6 is not allowed'}
        try:
            vpcs = self.conn.list_vpcs().result['vpcs']
            for vpc in vpcs:
                if self.vpc == vpc['name']:
                    vpc_id = vpc['id']
                    resource_group_id = vpc['resource_group']['id']
                    break
            else:
                return {'result': 'failure', 'reason': f'vpc {self.vpc} does not exist'}
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to retrieve vpc information. Hit {e}'}
        try:
            self.conn.create_subnet(vpc_v1.SubnetPrototypeSubnetByCIDR(
                name=name,
                ipv4_cidr_block=cidr,
                vpc=vpc_v1.VPCIdentityById(id=vpc_id),
                resource_group=vpc_v1.ResourceGroupIdentityById(id=resource_group_id),
                zone=vpc_v1.ZoneIdentityByHref(
                    href=get_zone_href(self.region, self.zone)
                ),
            ))
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to create network. Hit {e}'}
        return {'result': 'success'}

    def delete_network(self, name=None, cidr=None, force=False):
        try:
            subnets = self._get_subnets()
            for subnet in subnets:
                if name == subnet['name']:
                    subnet_id = subnet['id']
                    break
            else:
                return {'result': 'failure', 'reason': f'Subnet {name} not found'}
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to retrieve subnet {name} information. Hit {e}'}
        try:
            self.conn.delete_subnet(id=subnet_id)
        except ApiException as e:
            return {'result': 'failure', 'reason': f'Unable to delete subnet {name}. Hit {e}'}
        return {'result': 'success'}

    def list_pools(self):
        print("not implemented")
        return []

    def list_networks(self):
        networks = {}
        subnets = {}
        for subnet in self.conn.list_subnets().result['subnets']:
            newsubnet = {'name': subnet['name'], 'cidr': subnet['ipv4_cidr_block']}
            vpcid = subnet['vpc']['id']
            if vpcid in subnets:
                subnets[vpcid].append(newsubnet)
            else:
                subnets[vpcid] = [newsubnet]
        for net in self.conn.list_vpcs().result['vpcs']:
            networkname = net['name']
            vpcid = net['id']
            dhcp = net['default_network_acl']['name']
            mode = net['default_routing_table']['name']
            cidr = ''
            if vpcid in subnets:
                for subnet in subnets[vpcid]:
                    cidr = subnet['cidr']
                    if subnet['name'] == networkname:
                        break
            networks[networkname] = {'cidr': cidr, 'dhcp': dhcp, 'domain': vpcid, 'type': 'routed', 'mode': mode}
        return networks

    def info_network(self, name):
        networkinfo = common.info_network(self, name)
        return networkinfo

    def list_subnets(self):
        subnets = {}
        try:
            provisioned_subnets = self._get_subnets()
        except ApiException as e:
            error(f'Unable to retrieve subnets. Hit {e}')
            return subnets
        for subnet in provisioned_subnets:
            subnets[subnet['name']] = {
                'az': subnet['zone']['name'],
                'cidr': subnet['ipv4_cidr_block'],
                'network': subnet['vpc']['name']
            }
        return subnets

    def delete_pool(self, name, full=False):
        print("not implemented")

    def network_ports(self, name):
        return []

    def vm_ports(self, name):
        return []

    def get_pool_path(self, pool):
        print("not implemented")

    def list_flavors(self):
        flavors = []
        try:
            for flavor in self.conn.list_instance_profiles().result['profiles']:
                if self.debug:
                    print(flavor)
                flavors.append([flavor['name'], flavor['vcpu_count']['value'], flavor['memory']['value']])
            return flavors
        except ApiException as e:
            error(f"Unable to retrieve available flavors. Hit {e}")
            return []

    def export(self, name, image=None):
        print("not implemented")

    def _wait_lb_active(self, id):
        while True:
            result = self.conn.get_load_balancer(id=id).result
            if result['provisioning_status'] == 'active':
                break
            pprint("Waiting 10s for lb to go active...")
            sleep(10)

    def _wait_lb_dead(self, id):
        while True:
            try:
                self.conn.get_load_balancer(id=id).result
                pprint("Waiting 10s for lb to disappear...")
                sleep(10)
            except:
                break

    def create_loadbalancer(self, name, ports=[], checkpath='/index.html', vms=[], domain=None, checkport=80, alias=[],
                            internal=False, dnsclient=None, ip=None):

        ports = [int(port) for port in ports]
        internal = False if internal is None else internal
        clean_name = name.replace('.', '-')
        pprint(f"Creating Security Group {clean_name}")
        security_group_ports = ports + [int(checkport)] if int(checkport) not in ports else ports
        security_group_id = self.create_security_group(clean_name, {'ports': security_group_ports})
        subnets = set()
        member_list = []
        resource_group_id = None
        if vms:
            for vm in vms:
                try:
                    virtual_machine = self._get_vm(vm)
                except ApiException as e:
                    return {'result': 'failure', 'reason': f'Unable to retrieve VM {vm}. Hit {e}'}
                member_list.append(virtual_machine['primary_network_interface']['primary_ipv4_address'])
                if 'primary_network_interface' in virtual_machine:
                    subnets.add(virtual_machine['primary_network_interface']['subnet']['id'])
                nic_id = virtual_machine['primary_network_interface']['id']
                self.conn.add_security_group_network_interface(security_group_id, nic_id)
                if resource_group_id is None:
                    resource_group_id = virtual_machine['resource_group']['id']
                self.update_metadata(vm, 'loadbalancer', clean_name, append=True)
        pprint("Creating load balancer pool...")
        try:
            lb = self.conn.create_load_balancer(
                is_public=not internal,
                name=clean_name,
                pools=[vpc_v1.LoadBalancerPoolPrototype(
                    algorithm='round_robin',
                    health_monitor=vpc_v1.LoadBalancerPoolHealthMonitorPrototype(
                        delay=20,
                        max_retries=2,
                        timeout=3,
                        # type='http',
                        type='tcp',
                        url_path=checkpath,
                        port=checkport,
                    ),
                    protocol='tcp',
                    members=[vpc_v1.LoadBalancerPoolMemberPrototype(
                        port=port,
                        target=vpc_v1.LoadBalancerPoolMemberTargetPrototypeIP(address=m)
                    ) for m in member_list],
                    name=f"{clean_name}-{port}",
                ) for port in ports],
                subnets=[vpc_v1.SubnetIdentityById(id=x) for x in subnets],
                resource_group_id=vpc_v1.ResourceGroupIdentityById(id=resource_group_id),
                security_groups=[vpc_v1.SecurityGroupIdentityById(id=security_group_id)],
            ).result
            self._wait_lb_active(id=lb['id'])
        except ApiException as e:
            error(f'Unable to create load balancer. Hit {e}')
            return {'result': 'failure', 'reason': f'Unable to create load balancer. Hit {e}'}
        pprint("Creating listeners...")
        for index, port in enumerate(ports):
            try:
                self.conn.create_load_balancer_listener(
                    load_balancer_id=lb['id'],
                    port=port,
                    protocol='tcp',
                    default_pool=vpc_v1.LoadBalancerPoolIdentityById(id=lb['pools'][index]['id'])
                )
                try:
                    self._wait_lb_active(id=lb['id'])
                except ApiException as e:
                    error(f'Unable to create load balancer. Hit {e}')
                    return {'result': 'failure', 'reason': f'Unable to create load balancer. Hit {e}'}
            except ApiException as e:
                error(f'Unable to create load balancer listener. Hit {e}')
                return {'result': 'failure', 'reason': f'Unable to create load balancer listener. Hit {e}'}
        pprint(f"Load balancer DNS name {lb['hostname']}")
        resource_model = {'resource_id': lb['crn']}
        try:
            tag_names = [f'realname:{name}']
            if domain is not None:
                tag_names.append(f'domain:{domain}')
            if dnsclient is not None:
                tag_names.append(f'dnsclient:{dnsclient}')
            self.global_tagging_service.attach_tag(resources=[resource_model],
                                                   tag_names=tag_names,
                                                   tag_type='user')
        except ApiException as e:
            error(f'Unable to attach tags. Hit {e}')
            return {'result': 'failure', 'reason': f'Unable to attach tags. Hit {e}'}
        if domain is not None:
            while True:
                try:
                    result = self.conn.get_load_balancer(id=lb['id']).result
                except ApiException as e:
                    pprint(f'Unable to check load balancer ip. Hit {e}')
                    return {'result': 'failure', 'reason': f'Unable to check load balancer ip. Hit {e}'}
                if len(result['private_ips']) == 0:
                    pprint("Waiting 10s to get private ips assigned")
                    sleep(10)
                    continue
                break
            ip = result['public_ips'][0]['address']
            if dnsclient is not None:
                return ip
            self.reserve_dns(name, ip=ip, domain=domain, alias=alias)
        return {'result': 'success'}

    def delete_loadbalancer(self, name):
        domain = None
        dnsclient = None
        clean_name = name.replace('.', '-')
        try:
            lbs = {x['name']: x for x in self.conn.list_load_balancers().result['load_balancers']}
            if clean_name not in lbs:
                error(f'Load balancer {name} not found')
                return
            lb = lbs[clean_name]
        except ApiException as e:
            error(f'Unable to retrieve load balancers. Hit {e}')
            return
        try:
            tags = self.global_tagging_service.list_tags(attached_to=lb['crn']).result['items']
        except ApiException as e:
            error(f'Unable to retrieve tags. Hit {e}')
            return
        realname = name
        for tag in tags:
            tagname = tag['name']
            if tagname.count(':') == 1:
                key, value = tagname.split(':')
                if key == 'domain':
                    domain = value
                if key == 'dnsclient':
                    dnsclient = value
                if key == 'realname':
                    realname = value
        try:
            self.conn.delete_load_balancer(id=lb['id'])
        except ApiException as e:
            error(f'Unable to delete load balancer. Hit {e}')
            return
        if domain is not None and dnsclient is None:
            pprint(f"Deleting DNS {realname}.{domain}")
            self.delete_dns(realname, domain, name)
        self._wait_lb_dead(id=lb['id'])
        try:
            pprint(f"Deleting Security Group {clean_name}")
            self.delete_security_group(clean_name)
        except Exception as e:
            error(f'Unable to delete security group. Hit {e}')
        if dnsclient is not None:
            return dnsclient

    def list_loadbalancers(self):
        results = []
        try:
            lbs = self.conn.list_load_balancers().result['load_balancers']
        except ApiException as e:
            error(f'Unable to retrieve LoadBalancers. Hit {e}')
            return results
        if lbs:
            vms_by_addresses = {}
            for vm in self.conn.list_instances().get_result()['instances']:
                vms_by_addresses[vm['network_interfaces'][0]['primary_ipv4_address']] = vm['name']
        for lb in lbs:
            protocols = set()
            ports = []
            lb_id = lb['id']
            name = lb['name']
            ip = lb['hostname']
            try:
                listeners = self.conn.list_load_balancer_listeners(load_balancer_id=lb_id).result['listeners']
            except ApiException as e:
                error(f'Unable to retrieve listeners for load balancer {name}. Hit {e}')
                continue
            for listener in listeners:
                protocols.add(listener['protocol'])
                ports.append(str(listener['port']))
            target = []
            if 'pools' in lb:
                pool_id = lb['pools'][0]['id']
                pool = self.conn.get_load_balancer_pool(id=pool_id, load_balancer_id=lb_id).get_result()
                for member in pool['members']:
                    member_data = self.conn.get_load_balancer_pool_member(lb_id, pool_id, member['id']).get_result()
                    if member_data['target']['address'] in vms_by_addresses:
                        member_name = vms_by_addresses[member_data['target']['address']]
                        target.append(member_name)
            target = ','.join(target)
            results.append([name, ip, ','.join(protocols), '+'.join(ports), target])
        return results

    def create_bucket(self, bucket, public=False):
        if bucket in self.list_buckets():
            error(f"Bucket {bucket} already there")
            return
        # location = {'LocationConstraint': self.region}
        args = {'Bucket': bucket}  # , "CreateBucketConfiguration": location} #TODO: fix this.
        if public:
            args['ACL'] = 'public-read'
        self.s3.create_bucket(**args)

    def delete_bucket(self, bucket):
        if bucket not in self.list_buckets():
            error(f"Inexistent bucket {bucket}")
            return
        for obj in self.s3.list_objects(Bucket=bucket).get('Contents', []):
            key = obj['Key']
            pprint(f"Deleting object {key} from bucket {bucket}")
            self.s3.delete_object(Bucket=bucket, Key=key)
        self.s3.delete_bucket(Bucket=bucket)

    def delete_from_bucket(self, bucket, path):
        if bucket not in self.list_buckets():
            error(f"Inexistent bucket {bucket}")
            return
        self.s3.delete_object(Bucket=bucket, Key=path)

    def download_from_bucket(self, bucket, path):
        self.s3.download_file(bucket, path, path)

    def upload_to_bucket(self, bucket, path, overrides={}, temp_url=False, public=False):
        if not os.path.exists(path):
            error(f"Invalid path {path}")
            return None
        if bucket not in self.list_buckets():
            error(f"Bucket {bucket} doesn't exist")
            return None
        extra_args = {'Metadata': overrides} if overrides else {}
        if public:
            extra_args['ACL'] = 'public-read'
        dest = os.path.basename(path)
        with open(path, "rb") as f:
            self.s3.upload_fileobj(f, bucket, dest, ExtraArgs=extra_args)
        if temp_url:
            expiration = 600
            return self.s3.generate_presigned_url('get_object', Params={'Bucket': bucket, 'Key': dest},
                                                  ExpiresIn=expiration)
        return None

    def fast_upload_to_bucket(self, bucket, path):
        from ibm_s3transfer.aspera.manager import AsperaConfig, AsperaTransferManager
        transfer_manager = AsperaTransferManager(self.s3)
        ms_transfer_config = AsperaConfig(multi_session=2, multi_session_threshold_mb=100)
        transfer_manager = AsperaTransferManager(client=self.s3, transfer_config=ms_transfer_config)
        with AsperaTransferManager(self.s3) as transfer_manager:
            future = transfer_manager.upload(path, bucket, os.path.basename(path))
            future.result()

    def list_buckets(self):
        response = self.s3.list_buckets()
        return [bucket["Name"] for bucket in response['Buckets']]

    def list_bucketfiles(self, bucket):
        if bucket not in self.list_buckets():
            error(f"Inexistent bucket {bucket}")
            return []
        return [obj['Key'] for obj in self.s3.list_objects(Bucket=bucket).get('Contents', [])]

    def public_bucketfile_url(self, bucket, path):
        return f"https://s3.direct.{self.region}.cloud-object-storage.appdomain.cloud/{bucket}/{path}"

    def reserve_dns(self, name, nets=[], domain=None, ip=None, alias=[], force=False, primary=False, instanceid=None):
        if domain is None:
            domain = nets[0]
        pprint(f"Using domain {domain}...")
        cluster = None
        fqdn = f"{name}.{domain}"
        if fqdn.split('-')[0] == fqdn.split('.')[1]:
            cluster = fqdn.split('-')[0]
            name = '.'.join(fqdn.split('.')[:1])
            domain = fqdn.replace(f"{name}.", '').replace(f"{cluster}.", '')
        dnszone = self._get_dns_zone(domain)
        if dnszone is None:
            return
        if ip is None:
            counter = 0
            while counter != 100:
                ip = self.internalip(name)
                if ip is None:
                    sleep(5)
                    pprint(f"Waiting 5 seconds to grab internal ip and create DNS record for {name}...")
                    counter += 10
                else:
                    break
        if ip is None:
            error(f'Unable to find an IP for {name}')
            return
        try:
            dnszone.create_dns_record(name=name, type='A', ttl=60, content=ip)
        except ApiException as e:
            error(f'Unable to create DNS entry. Hit {e}')
            return
        if alias:
            for a in alias:
                if a == '*':
                    record_type = 'A'
                    content = ip
                    if cluster is not None and ('ctlplane' in name or 'worker' in name):
                        dnsname = f'*.apps.{cluster}.{domain}'
                    else:
                        dnsname = f'*.{name}.{domain}'
                else:
                    record_type = 'CNAME'
                    content = f"{name}.{domain}"
                    dnsname = f'{a}.{domain}' if '.' not in a else a
                try:
                    dnszone.create_dns_record(name=dnsname, type=record_type, ttl=60, content=content)
                except ApiException as e:
                    error(f'Unable to create DNS entry. Hit {e}')
                    return

    def create_dns(self):
        print("not implemented")

    def delete_dns(self, name, domain, allentries=False):
        dnszone = self._get_dns_zone(domain)
        if dnszone is None:
            return
        cluster = None
        fqdn = f"{name}.{domain}"
        if fqdn.split('-')[0] == fqdn.split('.')[1]:
            cluster = fqdn.split('-')[0]
            name = '.'.join(fqdn.split('.')[:1])
            domain = fqdn.replace(f"{name}.", '').replace(f"{cluster}.", '')
        dnsentry = name if cluster is None else f"{name}.{cluster}"
        entry = f"{dnsentry}.{domain}"
        clusterdomain = f"{cluster}.{domain}"
        try:
            records = dnszone.list_all_dns_records().get_result()['result']
        except ApiException as e:
            error(f"Unable to check DNS {dnszone['name']} records. Hit {e}")
            return
        recordsfound = False
        for record in records:
            if entry in record['name'] or ('ctlplane-0' in name and record['name'].endswith(clusterdomain))\
                    or (record['type'] == 'CNAME' and record['content'] == entry):
                record_identifier = record['id']
                try:
                    dnszone.delete_dns_record(dnsrecord_identifier=record_identifier)
                    recordsfound = True
                except ApiException as e:
                    error(f"Unable to delete record {record['name']}. Hit {e}")
        if not recordsfound:
            error(f"No records found for {entry}")
        return {'result': 'success'}

    def list_dns(self, domain):
        results = []
        dnszone = self._get_dns_zone(domain)
        if dnszone is None:
            return []
        try:
            records = dnszone.list_all_dns_records().get_result()['result']
        except ApiException as e:
            error(f"Unable to check DNS {dnszone['name']} records. Hit {e}")
            return results
        for record in records:
            ip = record['content']
            results.append([record['name'], record['type'], record['ttl'], ip])
        return results

    def _get_vm(self, name):
        result = self.conn.list_instances(name=name).result
        if result['total_count'] == 0:
            return None
        return result['instances'][0]

    def _get_vms(self):
        result = self.conn.list_instances().result
        if result['total_count'] == 0:
            return []
        return result['instances']

    def _get_subnet(self, name):
        subnets = self._get_subnets()
        for subnet in subnets:
            if name == subnet['name'] and subnet['zone']['name'] == self.zone:
                return subnet
        return None

    def _get_subnets(self):
        return self.conn.list_subnets().result['subnets']

    def _get_image(self, name):
        result = self.conn.list_images(name=name).result
        if len(result['images']) == 0:
            return None
        return result['images'][0]

    def _get_profiles(self):
        return {x['name']: x for x in self.conn.list_instance_profiles().result['profiles']}

    def _get_volumes(self):
        return {x['name']: x for x in self.conn.list_volumes().result['volumes']}

    def _get_dns_zone(self, domain):
        try:
            dnslist = self.dns.list_zones().get_result()['result']
        except ApiException as e:
            error(f'Unable to check DNS resources. Hit {e}')
            return None
        dnsfound = False
        for dnsresource in dnslist:
            dnsid = dnsresource['id']
            dnsname = dnsresource['name']
            if dnsname == domain:
                dnsfound = True
                break
        if not dnsfound:
            error(f'Domain {domain} not found')
            return None
        try:
            dnszone = DnsRecordsV1(authenticator=self.authenticator, crn=self.cis_resource_instance_id,
                                   zone_identifier=dnsid)
        except ApiException as e:
            error(f'Unable to check DNS zones for DNS {domain}. Hit {e}')
            return None
        return dnszone

    def create_security_group(self, name, overrides={}):
        ports = overrides.get('ports', [])
        vpc_id = [net['id'] for net in self.conn.list_vpcs().result['vpcs'] if net['name'] == self.vpc][0]
        vpc_identity_model = {'id': vpc_id}
        rules = []
        security_group_rule_prototype_model = {}
        security_group_rule_prototype_model['direction'] = 'outbound'
        security_group_rule_prototype_model['ip_version'] = 'ipv4'
        security_group_rule_prototype_model['protocol'] = 'all'
        security_group_rule_prototype_model['remote'] = {'cidr_block': '0.0.0.0/0'}
        rules.append(security_group_rule_prototype_model)
        for port in ports:
            if isinstance(port, str) or isinstance(port, int):
                protocol = 'tcp'
                fromport, toport = int(port), int(port)
            elif isinstance(port, dict):
                protocol = port.get('protocol', 'tcp')
                fromport = port.get('from')
                toport = port.get('to') or fromport
                if fromport is None:
                    warning(f"Missing from in {ports}. Skipping")
                    continue
            pprint(f"Adding rule from {fromport} to {toport} protocol {protocol}")
            security_group_rule_prototype_model = {}
            security_group_rule_prototype_model['direction'] = 'inbound'
            security_group_rule_prototype_model['ip_version'] = 'ipv4'
            security_group_rule_prototype_model['protocol'] = protocol
            security_group_rule_prototype_model['port_min'] = fromport
            security_group_rule_prototype_model['port_max'] = toport
            security_group_rule_prototype_model['remote'] = {'cidr_block': '0.0.0.0/0'}
            rules.append(security_group_rule_prototype_model)
        response = self.conn.create_security_group(vpc_identity_model, name=name, rules=rules).result
        return response['id']

    def delete_security_group(self, name):
        security_groups = self.conn.list_security_groups().result['security_groups']
        matching_sgs = [x for x in security_groups if x['name'] == name]
        if matching_sgs:
            security_group = matching_sgs[0]
            security_group_id = security_group['id']
            for n in security_group['network_interfaces']:
                self.conn.remove_security_group_network_interface(security_group_id, n['id'])
            self.conn.delete_security_group(security_group_id)

    def _add_sno_security_group(self, cluster):
        security_group_id = self.create_security_group(f"{cluster}-sno", {'ports': [80, 443, 6443]})
        vm = self._get_vm(f"{cluster}-ctlplane-0")
        nic_id = vm['primary_network_interface']['id']
        self.conn.add_security_group_network_interface(security_group_id, nic_id)

    def update_nic(self, name, index, network):
        print("not implemented")

    def update_network(self, name, dhcp=None, nat=None, domain=None, plan=None, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def list_security_groups(self, network=None):
        return [x['name'] for x in self.conn.list_security_groups().result['security_groups']]

    def update_security_group(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def info_subnet(self, name):
        print("not implemented")
        return {}

    def create_subnet(self, name, cidr, dhcp=True, nat=True, domain=None, plan='kvirt', overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def delete_subnet(self, name, force=False):
        print("not implemented")
        return {'result': 'success'}

    def update_subnet(self, name, overrides={}):
        print("not implemented")
        return {'result': 'success'}

    def list_dns_zones(self):
        return [d['name'] for d in self.dns.list_zones().get_result()['result']]

    def detach_disks(self, name):
        print("not implemented")
        return {'result': 'success'}
